import inspect
import functools
from typing import Dict, List, Any, Optional, Callable, Type

import logging
logger = logging.getLogger(__name__)


def issetdescriptor(object: Any) -> bool:
    """Return true if the object is a method descriptor with setters.

    But not if ismethod() or isclass() or isfunction() are true.
    """
    if inspect.isclass(object) or inspect.ismethod(object) or inspect.isfunction(object):
        # mutual exclusion
        return False
    tp = type(object)

    return hasattr(tp, "__set__")




def ismethod(object: Any) -> bool:
    """Check if the given object is a method.

    If the object is a class, return False.
    If the object is not a function or method, return False.

    :param object: The object to check.
    :return: True if the object is a method, False otherwise.
    """
    return inspect.ismethod(object) or (inspect.isfunction(object) and hasattr(object, '__self__'))

def has_method(cls: Type, methodName: str) -> bool:
    """
    Check if class cls has a method with name methodName directly,
    or in one of its super-classes.

    :param cls: The class to check.
    :param methodName: The name of the method to look for.
    :return: True if the method exists, False otherwise.
    """
    return inspect.isroutine(getattr(cls, methodName, None))


def update_function_defaults(func: Callable, new_defaults: Optional[Dict[str, Any]] = None,
                             remove_defaults: Optional[List[str]] = None, is_forced: bool = False):
    """
    Decorator to change and remove default values.
    See https://alex-ber.medium.com/my-inspect-module-aa2d311246cb for more details.

    :param func: function/method to apply on.
    :param new_defaults: Dictionary of new default values to apply.
    :param remove_defaults: List of parameter names to remove default values from.
    :param is_forced: If False (default), only func's params that do have default value will be updated.
                      If True, all func's params will be updated that are mentioned by name in new_defaults and remove_defaults.
    :return: Decorated function with modified default values.
    """
    if new_defaults is None:
        new_defaults = {}
    if remove_defaults is None:
        remove_defaults = []

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Get the signature of the original function
        sig = inspect.signature(func)

        # Create a new signature with updated default values
        new_params = []
        for param in sig.parameters.values():
            if is_forced or param.default is not inspect.Parameter.empty:
                if param.name in new_defaults:
                    new_param = param.replace(default=new_defaults[param.name])
                elif param.name in remove_defaults:
                    new_param = param.replace(default=inspect.Parameter.empty)
                else:
                    new_param = param
            else:
                new_param = param
            new_params.append(new_param)

        new_sig = sig.replace(parameters=new_params)

        # Bind the arguments to the new signature
        bound_args = new_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        # Call the original function with the bound arguments
        _bounded_args = bound_args.args
        _bounded_kwargs = bound_args.kwargs
        return func(*_bounded_args, **_bounded_kwargs)

    return wrapper


def resolve_function_args(func: Callable, *args: Any, **kwargs: Any) -> Dict[str, Any]:
    """
    Map both explicit and default arguments of a function call by parameter name.
    It merges args and kwargs, taking into account default values from func.
    See https://alex-ber.medium.com/my-inspect-module-aa2d311246cb for more details.

    Parameters:
    - func: The function whose arguments are to be mapped.
    - args: Positional arguments passed to the function.
    - kwargs: Keyword arguments passed to the function.

    Returns:
    - A dictionary mapping parameter names to their corresponding values.
    """
    sig = inspect.signature(func)
    bound_args = sig.bind_partial(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args.arguments




import logging

import os as _os
import sys as _sys
import contextlib
from collections import deque
from .parsers import is_empty
from .importer import importer

logger = logging.getLogger(__name__)


# inspider by https://github.com/theskumar/python-dotenv/blob/12439aac2d7c98b5acaf51ae3044b1659dc086ae/src/dotenv/main.py#L250
def _is_interactive(module):
    """ Decide whether this is running in a REPL or IPython notebook """
    return not hasattr(module, '__file__')


def fixabscwd():
    """
    See here https://medium.com/@alex_ber/making-relative-path-to-file-to-work-d5d0f1da67bf for documentation.
    """
    logger.info(f"cwd is {_os.getcwd()}")

    main_module = _sys.modules['__main__']

    if _is_interactive(main_module) or getattr(_sys, 'frozen', False):
        # Should work without __file__, e.g. in REPL or IPython notebook.
        pass
    else:
        main_dir = _os.path.dirname(_os.path.realpath(main_module.__file__))

        logger.info(f"Going to change os.chdir('{main_dir}')")

        _os.chdir(main_dir)

def path():
    """
    For older Python version uses`importlib_resources` module.
    For newer version built in `importlib.resources`.
    :return:
    """
    if _sys.version_info >= (3, 7):
        from importlib.resources import path as _path
    else:
        try:
            from importlib_resources import path as _path
        except ImportError:
            import warnings

            warning = (
                "You appear to be missing some optional dependencies (importlib_resources);"
            )
            warnings.warn(warning, ImportWarning)
            raise
    return _path


def load_env(**kwargs):
    """
    Convenient method from loading environment variables
    inside packed format (eggs, etc).

    if dotenv_path or stream is present it will be used.
    if ENV_PCK is present, dotenv_path will be constructed from ENV_PCK and ENV_NAME.
    Otherwise, kwargs will be forwarded as is to load_dotenv.


    Implementaion note:
    If available it uses importlib.resources API,
    if not it assumes existence of backport of importlib_resources.

    :param ENV_PCK: package where to find .env file Optional.
    :param ENV_NAME: Name of .env file. Optional.
    :param dotenv_path:  absolute or relative path to .env file. Optional.
    :param stream: `StringIO` object with .env content. Optional.
    :return:
    """

    # os.environ['AWS_DEFAULT_REGION'] = 'eu-central-1'
    try:
        from dotenv import load_dotenv
    except ImportError:
        import warnings


        warning = (
            "You appear to be missing some optional dependencies (python-dotenv);"
            "please 'python -m pip install alex-ber-utils[python-dotenv]'."
        )

        warnings.warn(warning, ImportWarning)
        raise

    l_path = path()

    dotenv_path = kwargs.get('dotenv_path', None)
    stream = kwargs.get('stream', None)
    ENV_PCK = kwargs.pop('ENV_PCK', None)

    if not is_empty(dotenv_path) or not is_empty(stream) or \
            (is_empty(dotenv_path) and is_empty(stream) and is_empty(ENV_PCK)):
        load_dotenv(**kwargs)

    else:
        if is_empty(ENV_PCK):
            raise ValueError("ENV_PCK can't be empty")

        ENV_NAME = kwargs.pop('ENV_NAME', '.env')
        with l_path(ENV_PCK, ENV_NAME) as full_path:
            d = {**kwargs, 'dotenv_path': full_path}

            load_dotenv(**d)



class BaseOsEnvrion:
    DEFAULT_DELIMSEP = _os.pathsep  # ';'
    DEFAULT_ENVSEP = _os.path.sep  # /
    DEFAULT_KEYSEP = ','

    def __init__(self, **kwargs):
        delimsep = kwargs.get('ENV_DELIM_SEP', None)
        if is_empty(delimsep):
            delimsep = self.DEFAULT_DELIMSEP
        self.delimsep = delimsep

        envsep = kwargs.get('ENV_SEP', None)
        if is_empty(envsep):
            envsep = self.DEFAULT_ENVSEP
        self.envsep = envsep


        keysep = kwargs.get('ENV_KEY_SEP', None)
        if is_empty(keysep):
            keysep = self.DEFAULT_KEYSEP
        self.keysep = keysep

        env_keys = kwargs.get('ENV_KEYS', None) # 'KEY'
        if is_empty(env_keys):
            raise ValueError('ENV_KEYS is not found in kwargs')
        self.env_keys = self._str_to_list(self.keysep, env_keys)

    def _str_to_list(self, sep, value):
        ret = [w.strip() for w in value.split(sep)]
        return ret

    def _list_to_str(self, sep, *args):
        ret = sep.join([*args])
        return ret


class OsEnvrionPathExpender(BaseOsEnvrion):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        pck = kwargs.get('ENV_MAIN_PCK', None)
        if is_empty(pck):
            raise ValueError('ENV_MAIN_PCK is not found in kwargs')

        self.pck = pck

    def fix_abs_path(self):
        l_path = path()

        with l_path(self.pck, '__init__.py') as init_file:
            full_prefix = str(init_file.parent.parent)

            for env_key in self.env_keys:
                rel_paths = self._str_to_list(self.delimsep, _os.environ[env_key])

                buffersize = 0 if rel_paths is None else len(rel_paths)
                env_values = deque(maxlen=buffersize)

                for rel_path in rel_paths:
                    abs_path = self._list_to_str(self.envsep, *[full_prefix, rel_path])
                    env_values.append(abs_path)

                values = self._list_to_str(self.delimsep, *env_values)
                _os.environ[env_key] = values


class OsEnvrionPathRetry(BaseOsEnvrion):

    def fix_retry_path(self):
        from pathlib import Path

        for env_key in self.env_keys:
            env_paths = self._str_to_list(self.delimsep, _os.environ[env_key])

            buffersize = 0 if env_paths is None else len(env_paths)
            env_values = deque(maxlen=buffersize)

            for env_path in env_paths:
                is_too_short = len(env_path) == 1 and env_path != '/' and env_path != '\\'
                if is_too_short:
                    continue

                env_path_p = Path(env_path)
                is_exists = env_path_p.exists()

                if not is_exists:
                    drive = env_path_p.drive
                    if len(drive) == 2 and drive[1] == ':':
                        env_path_p = env_path_p.as_posix()[2:]

                env_values.append(str(env_path_p))

            values = self._list_to_str(self.delimsep, *env_values)
            _os.environ[env_key] = values



def fix_env(**kwargs):
    """
    This method "fixes" os.environ relatively to the given ENV_MAIN_PCK package.
    This method ignores current working directory or __main__ module's __file__ atrribute.
    For each key in ENV_KEYS, this method prepends full_prefix to os.environ[key].
    full_prefix is calculated as absolute path of __init__.py of ENV_MAIN_PCK.



    :param ENV_KEYS keys of os.environ which will be fixed.
    :param ENV_MAIN_PCK: package to calcualte full_prefix.
    :param ENV_KEY_SEP: seperator used in ENV_PCK. Optional. Default is ','
    :param ENV_SEP: seperator that is used inside path. Optional. Default is os.path.sep.
    :param ENV_DELIM_SEP: seperator that is used inside os.environ. Optional. Default is os.pathsep.
    :param cls: class or str with implementation logic. Optional. Default is OsEnvrionPathExpender.
    """

    cls = kwargs.pop('cls', OsEnvrionPathExpender)
    if isinstance(cls, str):
        cls = importer(cls)
    expender = cls(**kwargs)
    expender.fix_abs_path()


def fix_retry_env(**kwargs):
    """
    This method "fixes" os.environ making path to Work both on Windows and Linux.

    For each key in ENV_KEYS, this method check if the value can be successfully resolved
    as path. If so, it does nothing. Otherwise, it strip away the drive part (i.e. C:\\)
    changing os.environ entry.


    :param ENV_KEYS keys of os.environ which will be "fixed".
    :param ENV_KEY_SEP: seperator used in ENV_PCK. Optional. Default is ','
    :param ENV_SEP: seperator that is used inside path. Optional. Default is os.path.sep.
    :param ENV_DELIM_SEP: seperator that is used inside os.environ. Optional. Default is os.pathsep.
    :param cls: class or str with implementation logic. Optional. Default is OsEnvrionPathRetry.
    """

    cls = kwargs.pop('cls', OsEnvrionPathRetry)
    if isinstance(cls, str):
        cls = importer(cls)
    path_retry = cls(**kwargs)
    path_retry.fix_retry_path()


@contextlib.contextmanager
def FixRelCwd(relPackage, logger=None):
    """
    This context-manager temporary changes current working directory to the one where relPackage is installed.
    If relPackage is Python script (or main method of entire application) that relies on relative path,
    for example, in order to get some resource from the file-system (for example, some configuration file,
    that lies in the same directory as relPackage) and you want to call relPackage from another directory, this
    context-manager makes relPackage 'just work'.

    :param relPackage: - installed package
    :param logger: - optionally, logger to be used, if not specified, default one will be used.
    :return:
    """
    if logger is None:
        logger = logging.getLogger(__name__)
    logger.debug(f"cwd is {_os.getcwd()}")

    main_module = _sys.modules['__main__']

    if _is_interactive(main_module) or getattr(_sys, 'frozen', False):
        # Should work without __file__, e.g. in REPL or IPython notebook.
        pass
    else:
        try:
            cwd = _os.getcwd()
            dir = _os.path.dirname(_os.path.realpath(relPackage.__file__))
            logger.debug(f"Going to change os.chdir('{dir}')")
            _os.chdir(dir)
            yield dir
        finally:
             _os.chdir(cwd)

@contextlib.contextmanager
def GuardedWorkerException(logger=None, suppress=False, default_exc_message="Worker failed"):
    """
    surround with try-except because of worker exceptions cause the pool.join() to halt forever - and thus memory leak!

    It is very difficult if not impossible to pickle 'exceptions' back to the parent process.
    Simple ones work, but many others don't.
    For example, CalledProcessError is not pickable (my guess, this is because of stdout, stderr data-members).

    see https://stackoverflow.com/questions/15314189/python-multiprocessing-pool-hangs-at-join
    see https://bugs.python.org/issue9400

    This bug still exists in Python 3.

    :param logger to log the exception. If None, exception will be logged to sys.stderr.
    :param suppress whether to suppress exception. Default False - Exception with default_exc_message
    will be raised.
    :param default_exc_message. If suppress is False, this will be used as message for Exception.
    """
    try:
        yield None
    except Exception as e:
        if logger is None:
            import traceback
            print("Caught exception in worker proccess", file=_sys.stderr)
            traceback.print_exc(file=_sys.stderr)
        else:
            logger.error("Caught exception in worker proccess", exc_info=e)
        if not suppress:
            raise Exception(default_exc_message)



def is_iterable(value):
    """
    Check if a value is iterable, excluding None, str and bytes.

    Args:
        value: The value to check.

    Returns:
        bool: True if the value is iterable, False otherwise.
    """
    if value is None or isinstance(value, (str, bytes)):
        return False

    try:
        iter(value)
        return True
    except TypeError:
        return False

def is_mapping(value):
    """
    Check if a value is a mapping.

    Args:
        value: The value to check.

    Returns:
        bool: True if the value is a mapping, False otherwise.
    """
    try:
        return hasattr(value, 'items') and callable(value.items)
    except AttributeError:
        return False

def make_hashable(obj):
    """
    Recursively convert objects to hashable types.

    This function ensures that objects can be used as keys in dictionaries
    or elements in sets. It handles mappings, iterables, and objects that have
    a `__hash__` method or fall back to string representation for non-hashable objects.

    Args:
        obj: The object to be converted.

    Returns:
        A hashable representation of the object, or a HashableWrapper instance
        if the object is inherently non-hashable.
    """
    if is_mapping(obj):
        return frozenset((make_hashable(k), make_hashable(v)) for k, v in obj.items())
    elif is_iterable(obj):
        return tuple(make_hashable(e) for e in obj)
    elif hasattr(obj, '__hash__') and obj.__hash__:
        return obj
    else:
        return HashableWrapper(obj)

class HashableWrapper:
    """
    Wraps an object to make it hashable. This is a fallback for objects that
    do not natively support hashing.
    """
    def __init__(self, obj):
        self.obj = obj

    def __hash__(self):
        return hash(str(self.obj))

    def __eq__(self, other):
        return isinstance(other, HashableWrapper) and self.obj == other.obj

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.obj)})'




"""
This module effectively changes default values of the standard pprint.pprint module.
Intended usage:

instead of: from pprint import pprint
use:        from alexber.utils.pprint import pprint

You can utilize any publicly available functions from pprint, including *.

One of the main changes is to set sort_dicts=False.
From Python 3.7 onward, dictionary order of dict is guaranteed to be the insertion order.
This behavior was an implementation detail of CPython from version 3.6.
As a result, the order of keys is now deterministic, eliminating the need to sort keys.

The defaults are:

indent: 4,
width: 120,
depth: None,
stream: None,
compact: False,
sort_dicts: False,
underscore_numbers: False

If a value is absent in your Python version's pprint (for example, underscore_numbers was added only in Python 3.10),
and you are using a lower version (e.g., Python 3.9), this argument will be safely ignored.

See https://alex-ber.medium.com/my-pprint-module-f25a7b695e5f for more details.
"""


import logging
logger = logging.getLogger(__name__)

import inspect
import pprint as _pprint
from .inspects import update_function_defaults
import importlib

def _calc_pprint_new_params():
    # Define the default values
    default_values = {
        'indent': 4,
        'width': 120,
        'depth': None,
        'stream': None,
        'compact': False,
        'sort_dicts': False,
        'underscore_numbers': False
    }

    # Get the parameters of PrettyPrinter.__init__
    init_signature = inspect.signature(_pprint.PrettyPrinter.__init__)
    valid_params_d = init_signature.parameters #dict-like

    # Remove unsupported values from default_values
    default_values = {k: v for k, v in default_values.items() if k in valid_params_d}
    return default_values


_pprint_new_params = _calc_pprint_new_params()


# Dynamically define __all__ based on _pprint's __all__
__all__ = _pprint.__all__.copy()

# Dynamically update defaults for each item in pprint
this_module = importlib.import_module(__name__)
for name in __all__:
    _obj = getattr(_pprint, name)
    if inspect.isfunction(_obj):
        updated_func = update_function_defaults(_obj, new_defaults=_pprint_new_params)
        setattr(this_module, name, updated_func)
    elif inspect.isclass(_obj):
        # Create a new class dynamically
        class _cloned_obj(_obj):
            pass

        updated_init = update_function_defaults(_cloned_obj.__init__, new_defaults=_pprint_new_params)
        setattr(_cloned_obj, '__init__', updated_init)
        setattr(this_module, name, _cloned_obj)


"""
This module is usable in your deployment script. See also `fabs` module.

See here https://medium.com/analytics-vidhya/my-ymlparsers-module-88221edf16a6 for documentation.

This module depends on some 3-rd party dependencies, in order to use it, you should have installed first. To do it
run `python -m pip install alex-ber-utils[yml]`.
"""

from collections import OrderedDict
from pathlib import Path
from collections import deque
try:
    import alexber.utils.ymlparsers as ymlparsers
except ImportError:
    pass

from .init_app_conf import merge_list_value_in_dicts, conf
from .parsers import is_empty, parse_sys_args


def split_path(filename, split_dirname):
    """
    Split filename in 2 part parts by split_dirname.
    first_part will ends with split_dirname.
    second_part will start immediately after split_dirname.

    if split_dirname is not in filename or split_dirname is None, the behaviour is undefined.
    if split_dirname exists twice in the path, the last one will be used for splitting.

    :param filename: path to filename, can be relative or absolute.
    :param split_dirname: directory name in filename that will be used to split the path
    :return: (first_part, second_part) - such that first_part+second_part is absolute path.

    """
    if split_dirname is None:
        raise FileNotFoundError("can't split on None")

    parts = Path(filename).parts
    split_dirname = str(split_dirname)


    if split_dirname not in parts:
        raise FileNotFoundError(f"{split_dirname} is not found in {filename}")
    length = len(parts)

    use = False
    second_parts = deque(maxlen=length)
    first_parts = deque(maxlen=length)
    for part in reversed(parts):
        if part == split_dirname:
            use = True
        if use:
            first_parts.appendleft(part)
        else:
            second_parts.appendleft(part)


    return Path(*first_parts), Path(*second_parts)

def add_to_zip_copy_function(split_dirname=None, zf=None):
    """
    Factory method that returns closure that can be used as copy_function param in shutil.copytree()

    :param split_dirname: path from this directory and below will be used in archive.
    :param zf: zipfile.ZipFile
    :return:
    """
    def add_to_zip_file(src,dst):
        """
        Closure that can be used as copy_function param in shutil.copytree()
        shutil.copytree() is used to add from src to archive with entries evaluted according to split_dirname.

        :param src: soource file to use in entry in archive
        :param dst: ignored, see split_dirname in enclused function
        :return:
        """
        _, last_parts = split_path(src, split_dirname)
        dest_path = Path(split_dirname) / last_parts
        zf.write(str(src), str(dest_path))
    return add_to_zip_file

def load_config(argumentParser=None, args=None):
    """
    Simplified method for parsing yml configuration file with optionally overrided profiles only.
    See alexber.utils.init_app_conf.parse_config() for another variant.

    :param argumentParser with instruction how to interpret args. If None, the default one will be instantiated.
    :param args: if not None will be used instead of sys.argv
    :return:
    """
    if ymlparsers.HiYaPyCo.jinja2ctx is None:
        raise ValueError("ymlparsers.HiYaPyCo.jinja2ctx can't be None")

    params, sys_d = parse_sys_args(argumentParser, args)
    config_file = params.config_file
    full_path = Path(config_file).resolve()  # relative to cwd

    with ymlparsers.DisableVarSubst():
        default_d = ymlparsers.load([str(full_path)])

    profiles = merge_list_value_in_dicts(sys_d, default_d, conf.GENERAL_KEY, conf.PROFILES_KEY)
    b = is_empty(profiles)
    if not b:
        # merge to default_d
        general_d = default_d.setdefault(conf.GENERAL_KEY, OrderedDict())
        general_d[conf.PROFILES_KEY] = profiles
        return full_path, default_d




"""
See here https://medium.com/analytics-vidhya/how-to-write-easily-customizable-code-8b00b43406b2 for documentation.
"""
import importlib
import logging
import inspect
import pkgutil

logger = logging.getLogger(__name__)




#adopted from scrapy
def _walk_modules(path):
    """
    Loads a module and all its submodules from the given module path and
    returns them. If *any* module throws an exception while importing, that
    exception is thrown back.
    """

    # Support for namespace packages is added. See PEP 420.
    # Namespace packages are a mechanism for splitting a single Python package across multiple directories on disk.
    # When interpreted encounter with non-empty __path__ attribute it adds modules found in those locations
    # to the current package.

    mods = []
    mod = importlib.import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in pkgutil.iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += _walk_modules(fullpath)
            else:
                submod = importlib.import_module(fullpath)
                mods.append(submod)
    return mods


#adopted from mock.mock._dot_lookup
def _dot_lookup(thing, comp, import_path):
    '''
    Recursively import packages (if needed) by dotes.
    '''
    try:
        return getattr(thing, comp)
    except AttributeError:
        importlib.import_module(import_path)
        _walk_modules(import_path)
        return getattr(thing, comp)

#adopted from mock.mock._importer
def importer(target):
    '''
    Convert str to Python construct that target is represented.
    This method will recursively import packages (if needed)
    Following dot notation from left to right. If the component
    exists in packagage (is defined and imported) it will be used,
    otherwrise, it will be imported.

    This method supports PEP 420 (implicit Namespace Packages).

    Note: only compile-time construct is supported.
    Note: no instances will be returned from here, only classes.

    :param target: str to lookup
    :return: function/module/class, etc
    '''

    components = target.split('.')
    import_path = components.pop(0)
    thing = importlib.import_module(import_path)
    _walk_modules(import_path)

    for comp in components:
        import_path += f".{comp}"
        thing = _dot_lookup(thing, comp, import_path)
    return thing



def new_instance(target, *args, **kwargs):
    '''
    Convert str to Python construct that target is represented.
    This method will recursively import packages (if needed)
    Following dot notation from left to right. If the component
    exists in package (is defined and imported) it will be used,
    otherwrise, it will be imported.

    If target doesn't represent the class, it will be returned as is.
    If target is the class, instance of it will be created,
    args and kwargs will be passed in to appropriate
    __new__() / __init__() / __init_subclass__() methods.


    :param target:
    :param args:   - position args  for c-tor
    :param kwargs: - key-value args for c-tor
    :return:
    '''
    thing = importer(target)
    ret = thing
    if inspect.isclass(thing):
        ret = thing(*args, **kwargs)
    return ret


import logging
logger = logging.getLogger(__name__)
import pytest

from alexber.utils.props import Properties, lazyproperty
from importlib.resources import open_text


def test_parse_java_properties_alternative_delim(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch.object(Properties, 'PLACEHOLDER_TOKEN', new='@', spec_set=True)
    exp_cls_name_a = 'alexber.rpsgame.players.ConstantPlayer'
    exp_cls_name_b = exp_cls_name_a
    expdd = {'playera.cls': 'alexber.rpsgame.players.ConstantPlayer',
             'playerb.cls': 'alexber.rpsgame.players.ConstantPlayer', }


    p = Properties()
    pck = '.'.join(['tests_data', __package__])

    with open_text(pck, 'config2.properties') as f:
        p.load(f)
    dd = p.as_dict()

    assert expdd == dd


def test_parse_java_properties(request):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_cls_name_a = 'alexber.rpsgame.players.ConstantPlayer'
    exp_cls_name_b = exp_cls_name_a
    expdd = {'playera.cls': 'alexber.rpsgame.players.ConstantPlayer',
             'playerb.cls': 'alexber.rpsgame.players.ConstantPlayer',}

    p = Properties()
    pck = '.'.join(['tests_data', __package__])
    with open_text(pck, 'config.properties') as f:
        p.load(f)
    dd = p.as_dict()
    #we want to ignore key inner.*
    a_cls_name = dd['playera.cls']
    b_cls_name = dd['playerb.cls']

    d = {'playera.cls': a_cls_name,
         'playerb.cls': b_cls_name,}

    assert expdd==d

class ProperyBase:

    @property
    def answer(self):
        return -1

class ProperyDerived(ProperyBase):
    @property
    def answer(self):
        return 42

def test_regular_properties_support_override(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    a = ProperyDerived()
    actual = a.answer
    assert 42 == actual

class LazyProperty:
    def __init__(self, a=42):
        self._answer = a

    @lazyproperty
    def answer(self):
        return self._make_long_calc()





if __name__ == "__main__":
    pytest.main([__file__])



import logging
import threading
import time
import pytest
from smtplib import SMTP as _SMTP, LMTP as _LMTP
from alexber.utils.emails import SMTPHandler, OneMemoryHandler
import alexber.utils.emails as emails
#don't remove this
from platform import uname
from alexber.utils.emails import EmailStatus, FINISHED
logger = logging.getLogger(__name__)
emailLogger = None

@pytest.fixture
def errorMailHandler(mocker):

    d1 = {
     #   'level': logging.DEBUG,
        'smtpclsname': 'smtplib.SMTP',
        'mailhost': ['smtp-relay.gmail.com', 587],
        'fromaddr': f'{{uname()[1]}}@gmail.com',
        'toaddrs': ['no-reply@gmail.com'],
        'subject': 'mail_subject'
    }

    error_email_handler = SMTPHandler(**d1)
    error_email_handler.setLevel(logging.DEBUG)

    mocker.patch.object(error_email_handler, 'emit',autospec=True, spec_set=True)
    return error_email_handler


@pytest.fixture
def emailsSubject(request, mocker):
    request_param = 'Process Status : {{status}}' if (
        not hasattr(request, 'param')) else request.param
    yield request_param

@pytest.fixture
def errorMailMemoryHandler(request, mocker, errorMailHandler, emailsSubject):

    d2 = {
        #'formatter': detailFormatter,
        #'subject': 'Process Status : {{status}}',
        'subject': emailsSubject,
        #'target': error_email_handler
    }

    detailFormatter = logging.Formatter('%(asctime)-15s %(levelname)s [%(name)s.%(funcName)s] %(message)s',
                                        '%Y-%m-%d %H:%M:%S')
    error_mail_memory_handler = OneMemoryHandler(**d2)
    error_mail_memory_handler.setFormatter(detailFormatter)
    error_mail_memory_handler.setTarget(errorMailHandler)
    return error_mail_memory_handler

@pytest.fixture
def emailsFixture(request, mocker, errorMailMemoryHandler):
    global emailLogger
    emailLogger = logging.getLogger('emails')
    propagate_param = emailLogger.propagate if (not hasattr(request, 'param')) else request.param
    prev_propogate = emailLogger.propagate
    emailLogger.propagate = propagate_param
    emailLogger.addHandler(errorMailMemoryHandler)
    yield errorMailMemoryHandler.target
    emailLogger.removeHandler(errorMailMemoryHandler)
    emailLogger.propagate = prev_propogate

def run_successfuly():
    logger.debug("run_successfuly()")
    emailLogger.info("Start - Process")
    emailLogger.info("Step 1")
    emailLogger.info("Step 2")
    emailLogger.info("Done Successfully - Process")

def run_with_failure():
    logger.debug("run_with_failure()")
    emailLogger.info("Start - Process")
    emailLogger.info("Step 1")
    1/0
    emailLogger.info("Step 2")
    emailLogger.info("Done Successfully - Process")

def is_failed(logrecord):
    subject = str(logrecord.msg['subject'])
    assert '{status}' not in subject
    assert 'Done' in subject or 'Failed' in subject
    b = 'Failed' in subject
    return b

def check_failed(logrecord):
    subject = str(logrecord.msg['subject'])
    pytest.assume('Failed' in subject)
    pytest.assume('{status}' not in subject)
    message = str(logrecord.msg)
    pytest.assume('Start' in message)
    pytest.assume('division by zero' in message)
    pytest.assume('Step 1' in message)
    pytest.assume('Step 2' not in message)


def check_sucess(logrecord):
    subject = str(logrecord.msg['subject'])
    pytest.assume('Done' in subject)
    pytest.assume('{status}' not in subject)
    message = str(logrecord.msg)
    pytest.assume('Start' in message)
    pytest.assume('Step 1' in message)
    pytest.assume('Step 2' in message)

def test_emails_intented_success(request, mocker, emailsFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    with EmailStatus(emailLogger=emailLogger, logger=None, faildargs={'status': 'Failed'},
                     successargs={'status': 'Done'}):
        run_successfuly()
    mock_log = emailsFixture.emit
    pytest.assume(mock_log.call_count == 1)
    (logrecord,), _ = mock_log.call_args

    check_sucess(logrecord)

#emailsSubject
@pytest.mark.parametrize('emailsSubject', ['Aggregates log from the Demo application',
                                                    ], indirect=True)
def test_emails_intented_simple_success(request, mocker, emailsFixture, emailsSubject):
    logger.info(f'{request._pyfuncitem.name}()')

    emailLogger.log(FINISHED, None)
    mock_log = emailsFixture.emit
    pytest.assume(mock_log.call_count == 1)
    (logrecord,), _ = mock_log.call_args
    subject = str(logrecord.msg['subject'])
    pytest.assume(emailsSubject in subject)

#emailsSubject
@pytest.mark.parametrize('emailsSubject', ['Aggregates log from the Demo application',
                                                    ], indirect=True)
def test_emails_intented_simple_success2(request, mocker, emailsFixture, emailsSubject):
    logger.info(f'{request._pyfuncitem.name}()')

    emailLogger.log(FINISHED, '')
    mock_log = emailsFixture.emit
    pytest.assume(mock_log.call_count == 1)
    (logrecord,), _ = mock_log.call_args
    subject = str(logrecord.msg['subject'])
    pytest.assume(emailsSubject in subject)

#emailsSubject
@pytest.mark.parametrize('emailsSubject', ['Aggregates log from the Demo application',
                                                    ], indirect=True)
def test_emails_intented_simple_success_wrong(request, mocker, emailsFixture, emailsSubject):
    logger.info(f'{request._pyfuncitem.name}()')

    with pytest.raises(ValueError):
        emailLogger.log(FINISHED, ['some non-dict value'])

@pytest.fixture
def specialAbruptCleanup(request, mocker):
    yield None
    setattr(emails._thread_locals, 'buffer', [])


def test_emails_intented_abrupt_execution(request, mocker, emailsFixture, errorMailMemoryHandler, specialAbruptCleanup):
    logger.info(f'{request._pyfuncitem.name}()')
    emailLogger.info("Started")
    emailLogger.info('Step 1')
    logging.shutdown([lambda: errorMailMemoryHandler])
    emailLogger.info('Step 2')

    mock_log = emailsFixture.emit
    pytest.assume(mock_log.call_count == 1)
    (logrecord,), _ = mock_log.call_args

    subject = str(logrecord.msg['subject'])
    pytest.assume('Abrupt' in subject)
    pytest.assume('{status}' not in subject)
    message = str(logrecord.msg)
    pytest.assume('Start' in message)
    pytest.assume('Step 1' in message)
    pytest.assume('Step 2' not in message)


@pytest.mark.parametrize('emailsFixture', [False], indirect=True)
def test_emails_intented_failure(request, mocker, emailsFixture):
    logger.info(f'{request._pyfuncitem.name}()')


    with EmailStatus(emailLogger=emailLogger, logger=None, faildargs={'status': 'Failed'},
                     successargs={'status': 'Done'}):
        run_with_failure()
    mock_log = emailsFixture.emit
    pytest.assume(mock_log.call_count == 1)
    (logrecord,), _ = mock_log.call_args


    check_failed(logrecord)


def test_init_config_default(request, mocker, emailsFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    emails.initConfig()
    pytest.assume(_SMTP==emails.default_smtp_cls)


def test_init_config2(request, mocker, emailsFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_default_smtp_port='9000'
    emails.initConfig(default_smtp_cls_name='smtplib.LMTP')
    pytest.assume(_LMTP==emails.default_smtp_cls)


@pytest.mark.parametrize('emailsFixture', [False], indirect=True)
def test_emails_multithreaded(request, mocker, emailsFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    stop =10

    def _run_successfuly(stop):
        for i in range(stop):
            with EmailStatus(emailLogger=emailLogger, logger=None, faildargs={'status': 'Failed'},
                             successargs={'status': 'Done'}):
                run_successfuly()
        time.sleep(1)

    def _run_with_failure(stop):
        for i in range(stop):
            with EmailStatus(emailLogger=emailLogger, logger=None, faildargs={'status': 'Failed'},
                             successargs={'status': 'Done'}):
                run_with_failure()


    th1 = threading.Thread(name="run_successfuly",
                               target=_run_successfuly, args=(stop,))

    th2 = threading.Thread(name="run_with_failure",
                           target=_run_with_failure, args=(stop,))

    th1.start()
    th2.start()
    th1.join()
    th2.join()

    mock_log = emailsFixture.emit
    pytest.assume(mock_log.call_count == 2*stop)
    for kall in mock_log.call_args_list:
        (logrecord,), _ = kall
        if is_failed(logrecord):
            check_failed(logrecord)
        else:
            check_sucess(logrecord)





if __name__ == "__main__":
    pytest.main([__file__])



import concurrent.futures

import logging
import threading
import types
import asyncio
from concurrent.futures import ThreadPoolExecutor
import contextvars
import pytest
import pytest
import asyncio
#import time


from alexber.utils.thread_locals import RLock, LockingCallableMixin, \
    LockingIterableMixin, LockingIterator, LockingAsyncIterableMixin, LockingAsyncIterator, LockingAccessMixin, \
    LockingPedanticObjMixin, LockingDefaultLockMixin, _coerce_base_language_model, LockingBaseLanguageModelMixin, \
    _is_pydantic_obj
from alexber.utils.thread_locals import threadlocal_var, get_threadlocal_var, del_threadlocal_var
from alexber.utils.thread_locals import exec_in_executor, exec_in_executor_threading_future, \
                                        get_main_event_loop


logger = logging.getLogger(__name__)


def test_get_threadlocal_var_empty(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


def test_get_threadlocal_var_exist(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    expValue = 1
    thread_locals.value = expValue

    value = get_threadlocal_var(thread_locals, 'value')
    assert expValue == value

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


ns = threading.local()
stop = 10

class Worker(threading.Thread):

    def run(self):
        w_logger = logging.getLogger(self.name)
        i = 0
        ns.val = 0

        for i in range(stop):
            ns.val += 1
            i+=1
            w_logger.debug(f"Thread: {self.name}, value: {ns.val}")
            value = get_threadlocal_var(ns, "val")
            assert i == value
        value = get_threadlocal_var(ns, "val")
        assert stop==value


def test_get_threadlocal_var_exist_different_thread(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    w1 = Worker()
    w2 = Worker()
    w1.start()
    w2.start()
    w1.join()
    w2.join()

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'val')

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'nonexist')


def test_del_threadlocal_var_empy(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    del_threadlocal_var(thread_locals, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


def test_del_threadlocal_var_exist(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    thread_locals.value = 1

    del_threadlocal_var(thread_locals, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')

class Box(object):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


def test_threadlocal_var_empty(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    expValue = 5
    box = threadlocal_var(thread_locals, 'value', Box, value=expValue)
    value = box.value
    assert expValue == value

    box2 = get_threadlocal_var(thread_locals, 'value')
    value2 = box2.value
    assert expValue == value2

    assert box == box2

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


def test_threadlocal_var_exists(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    expValue = 5
    box = threadlocal_var(thread_locals, 'value', Box, value=expValue)
    value = box.value
    assert expValue == value

    box2= threadlocal_var(thread_locals, 'value', Box, value=100)
    value2 = box2.value
    assert expValue == value2

    assert box == box2

    box3 = threadlocal_var(thread_locals, 'value', Box)
    value3 = box3.value
    assert expValue == value3

    assert box == box3

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')

async def example_task(return_value):
    return return_value

def create_coroutine_mock(mocker, return_value=None):
    async def mock_coroutine(*args, **kwargs):
        return return_value

    mock = mocker.AsyncMock(wraps=mock_coroutine)
    mock.__code__ = mock_coroutine.__code__
    mock.__class__ = types.FunctionType
    return mock

### Basic Tests for Synchronous Locks
def test_sync_acquire_release(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    assert lock.acquire() is True
    assert lock._sync_owner == threading.current_thread()
    assert lock._sync_count == 1
    lock.release()
    assert lock._sync_owner is None
    assert lock._sync_count == 0

def test_sync_reentrant(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    assert lock.acquire() is True
    assert lock.acquire() is True
    assert lock._sync_owner == threading.current_thread()
    assert lock._sync_count == 2
    lock.release()
    assert lock._sync_count == 1
    lock.release()
    assert lock._sync_owner is None
    assert lock._sync_count == 0

@pytest.mark.asyncio
async def test_async_acquire_release(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    assert await lock.async_acquire() is True
    assert lock._async_owner == asyncio.current_task()
    assert lock._async_count == 1
    await lock.async_release()
    assert lock._async_owner is None
    assert lock._async_count == 0

@pytest.mark.asyncio
async def test_async_reentrant(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    assert await lock.async_acquire() is True
    assert await lock.async_acquire() is True
    assert lock._async_owner == asyncio.current_task()
    assert lock._async_count == 2
    await lock.async_release()
    assert lock._async_count == 1
    await lock.async_release()
    assert lock._async_owner is None
    assert lock._async_count == 0

def test_sync_context_manager(request, mocker):
    lock = RLock()
    with lock:
        assert lock._sync_owner == threading.current_thread()
        assert lock._sync_count == 1
    assert lock._sync_owner is None
    assert lock._sync_count == 0

@pytest.mark.asyncio
async def test_async_context_manager(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    async with lock:
        assert lock._async_owner == asyncio.current_task()
        assert lock._async_count == 1
    assert lock._async_owner is None
    assert lock._async_count == 0

### Additional Fairness Tests
def test_sync_fairness(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    results = []

    def thread_func(thread_id):
        lock.acquire()
        results.append(thread_id)
        lock.release()

    threads = [threading.Thread(target=thread_func, args=(i,)) for i in range(3)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert results == [0, 1, 2], f"Results were {results}"

@pytest.mark.asyncio
async def test_async_fairness(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    lock = RLock()
    results = []

    async def async_task(task_id):
        await lock.async_acquire()
        results.append(task_id)
        await lock.async_release()

    tasks = [asyncio.create_task(async_task(i)) for i in range(3)]
    await asyncio.gather(*tasks)

    assert results == [0, 1, 2], f"Results were {results}"

### Other Callable and Iterable Tests for Completeness
def test_call_synchronous_function(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = mocker.Mock(return_value="result")
    lock = mocker.Mock()
    lock.__enter__ = mocker.Mock(return_value=None)
    lock.__exit__ = mocker.Mock(return_value=None)
    mixin = LockingCallableMixin(obj=obj, lock=lock)
    result = mixin()
    assert result == "result"
    obj.assert_called_once()
    lock.__enter__.assert_called_once()
    lock.__exit__.assert_called_once()

@pytest.mark.asyncio
async def test_locking_call_asynchronous_function(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = create_coroutine_mock(mocker, return_value="result")
    lock = mocker.AsyncMock()
    lock.__aenter__ = mocker.AsyncMock(return_value=None)
    lock.__aexit__ = mocker.AsyncMock(return_value=None)
    mixin = LockingCallableMixin(obj=obj, lock=lock)
    result = await mixin()
    assert result == "result"
    obj.assert_awaited_once()
    lock.__aenter__.assert_awaited_once()
    lock.__aexit__.assert_awaited_once()

def test_locking_iterable_mixin(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = [1, 2, 3]
    lock = mocker.Mock()
    lock.__enter__ = mocker.Mock(return_value=None)
    lock.__exit__ = mocker.Mock(return_value=None)
    mixin = LockingIterableMixin(obj=obj, lock=lock)
    iterator = iter(mixin)
    assert isinstance(iterator, LockingIterator)

    # Iterate over the iterator once
    result = list(iterator)
    assert result == [1, 2, 3]

    # Verify the exact number of calls to __enter__ and __exit__
    assert lock.__enter__.call_count == 3 + 1   # 1 for iterator exhaustion
    assert lock.__exit__.call_count == 3 + 1    # 1 for iterator exhaustion

def test_locking_iterator(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = iter([1, 2, 3])
    lock = mocker.Mock()
    lock.__enter__ = mocker.Mock(return_value=None)
    lock.__exit__ = mocker.Mock(return_value=None)
    iterator = LockingIterator(obj, lock)
    assert next(iterator) == 1
    assert next(iterator) == 2
    assert next(iterator) == 3
    with pytest.raises(StopIteration):
        next(iterator)
    assert lock.__enter__.call_count == 3 + 1   # 1 for iterator exhaustion
    assert lock.__exit__.call_count == 3 + 1    # 1 for iterator exhaustion

@pytest.mark.asyncio
async def test_locking_async_iterable_mixin(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    async def async_gen():
        for i in range(3):
            yield i

    obj = async_gen()
    lock = mocker.AsyncMock()
    lock.__aenter__ = mocker.AsyncMock(return_value=None)
    lock.__aexit__ = mocker.AsyncMock(return_value=None)
    mixin = LockingAsyncIterableMixin(obj=obj, lock=lock)
    async_iterator = mixin.__aiter__()
    assert isinstance(async_iterator, LockingAsyncIterator)

    result = []
    async for item in async_iterator:
        result.append(item)
    assert result == [0, 1, 2]

    # Verify the exact number of calls to __aenter__ and __aexit__
    assert lock.__aenter__.call_count == 3 + 1   # 1 for iterator exhaustion
    assert lock.__aexit__.call_count == 3 + 1    # 1 for iterator exhaustion

@pytest.mark.asyncio
async def test_locking_async_iterator(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    async def async_gen():
        for i in range(3):
            yield i

    obj = async_gen()
    lock = mocker.AsyncMock()
    lock.__aenter__ = mocker.AsyncMock(return_value=None)
    lock.__aexit__ = mocker.AsyncMock(return_value=None)
    async_iterator = LockingAsyncIterator(obj, lock)

    result = []
    async for item in async_iterator:
        result.append(item)
    assert result == [0, 1, 2]

    # Verify the exact number of calls to __aenter__ and __aexit__
    assert lock.__aenter__.call_count == 3 + 1   # 1 for iterator exhaustion
    assert lock.__aexit__.call_count == 3 + 1    # 1 for iterator exhaustion

def test_property_locking_access_mixin(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    class TestClass:
        @property
        def prop(self):
            return "value"

    obj = TestClass()
    lock = mocker.Mock()
    mixin = LockingAccessMixin(obj=obj, lock=lock)
    assert mixin.prop == "value"

def test_locking_access_sync_method(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = mocker.Mock()
    lock = mocker.Mock()
    lock.__enter__ = mocker.Mock(return_value=None)
    lock.__exit__ = mocker.Mock(return_value=None)

    mixin = LockingAccessMixin(obj=obj, lock=lock)

    def sync_method():
        return "result"

    obj.sync_method = sync_method
    wrapped_method = mixin.sync_method
    result = wrapped_method()
    assert result == "result"
    lock.__enter__.assert_called_once()
    lock.__exit__.assert_called_once()

@pytest.mark.asyncio
async def test_locking_access_async_method(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = mocker.Mock()
    lock = mocker.AsyncMock()
    lock.__aenter__ = mocker.AsyncMock(return_value=None)
    lock.__aexit__ = mocker.AsyncMock(return_value=None)

    mixin = LockingAccessMixin(obj=obj, lock=lock)

    async def async_method():
        return "result"

    obj.async_method = async_method
    wrapped_method = mixin.async_method()
    result = await wrapped_method
    assert result == "result"
    lock.__aenter__.assert_called_once()
    lock.__aexit__.assert_called_once()

def test_locking_access_property_handling(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    class TestClass:
        @property
        def prop(self):
            return "value"

    obj = TestClass()
    lock = mocker.Mock()
    lock.__enter__ = mocker.Mock(return_value=None)
    lock.__exit__ = mocker.Mock(return_value=None)

    mixin = LockingAccessMixin(obj=obj, lock=lock)
    assert mixin.prop == "value"

def test_locking_access_special_case_for_pydantic(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_pydantic_obj', return_value=True)

    class PydanticModel:
        def _copy_and_set_values(self):
            pass

    obj = PydanticModel()
    lock = mocker.Mock()
    lock.__enter__ = mocker.Mock(return_value=None)
    lock.__exit__ = mocker.Mock(return_value=None)

    mixin = LockingAccessMixin(obj=obj, lock=lock)
    wrapped_method = mixin._copy_and_set_values
    result = wrapped_method()
    assert result == mixin
    lock.__enter__.assert_called_once()
    lock.__exit__.assert_called_once()

def test_locking_default_lock_initialization_with_provided_lock(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = mocker.Mock()
    lock = mocker.Mock()
    mixin = LockingDefaultLockMixin(obj=obj, lock=lock)
    assert mixin._lock == lock

def test_locking_default_lock_initialization_without_provided_lock(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    obj = mocker.Mock()
    mock_rlock = mocker.patch('alexber.utils.thread_locals.RLock', autospec=True)
    mixin = LockingDefaultLockMixin(obj=obj)
    assert isinstance(mixin._lock, RLock)
    mock_rlock.assert_called_once()

def test_coerce_base_language_model_not_available(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_available_base_language_model', False)

    obj = mocker.Mock()
    _coerce_base_language_model(obj)
    # Ensure that no further action is taken when the model is not available
    # Since we are mocking, we should check that BaseLanguageModel.register is not called
    assert not mocker.patch('alexber.utils.thread_locals.BaseLanguageModel').register.called

def test_coerce_base_language_model_with_base_language_model(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_available_base_language_model', True)

    class MockBaseLanguageModel:
        @staticmethod
        def register(cls):
            pass

    mocker.patch('alexber.utils.thread_locals.BaseLanguageModel', MockBaseLanguageModel)

    class TestModel:
        _obj = MockBaseLanguageModel()

    obj = TestModel()
    mock_register = mocker.patch.object(MockBaseLanguageModel, 'register')
    _coerce_base_language_model(obj)
    mock_register.assert_called_once_with(type(obj))

def test_coerce_base_language_model_with_non_base_language_model(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_available_base_language_model', True)

    class MockBaseLanguageModel:
        @staticmethod
        def register(cls):
            pass

    mocker.patch('alexber.utils.thread_locals.BaseLanguageModel', MockBaseLanguageModel)

    class TestModel:
        _obj = mocker.Mock()

    proxy = TestModel()
    mock_register = mocker.patch.object(MockBaseLanguageModel, 'register')
    _coerce_base_language_model(proxy)
    mock_register.assert_not_called()

def test_locking_base_language_model_mixin_calls_coerce(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    # Mock the availability flag
    mocker.patch('alexber.utils.thread_locals._is_available_base_language_model', True)

    # Mock the BaseLanguageModel as a type
    class MockBaseLanguageModel:
        @staticmethod
        def register(cls):
            pass

    mocker.patch('alexber.utils.thread_locals.BaseLanguageModel', MockBaseLanguageModel)

    # Mock the _coerce_base_language_model function
    mock_coerce = mocker.patch('alexber.utils.thread_locals._coerce_base_language_model')
    class MockRootMixin:
        pass
    class LockingBaseLanguageModelMixin(MockRootMixin, MockBaseLanguageModel):
        def __init__(self, **kwargs):
            self._obj = kwargs.get('obj')
            mock_coerce(self)
    obj = mocker.Mock()
    instance = LockingBaseLanguageModelMixin(obj=obj)
    mock_coerce.assert_called_once_with(instance)

def test_coerce_base_language_model_checks_proxy_obj(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_available_base_language_model', True)
    class MockBaseLanguageModel:
        @staticmethod
        def register(cls):
            pass
    mocker.patch('alexber.utils.thread_locals.BaseLanguageModel', MockBaseLanguageModel)
    class TestModel:
        _obj = MockBaseLanguageModel()
    proxy = TestModel()
    mock_register = mocker.patch.object(MockBaseLanguageModel, 'register')
    _coerce_base_language_model(proxy)
    mock_register.assert_called_once_with(type(proxy))
    # Ensure that the function checks proxy._obj and not just obj
    assert isinstance(proxy._obj, MockBaseLanguageModel)

# Define a mock BaseModel class
class MockBaseModel:
    pass

@pytest.fixture
def mocker_pydantic(mocker):
    # Mock the availability of pydantic v1 and v2
    mocker.patch('alexber.utils.thread_locals._is_available_pydantic_v1', True)
    mocker.patch('alexber.utils.thread_locals._is_available_pydantic_v2', True)
    # Mock the import of pydantic and its BaseModel
    mocker.patch.dict('sys.modules', {
        'pydantic': mocker.MagicMock(BaseModel=MockBaseModel),
        'pydantic.v1': mocker.MagicMock(BaseModel=MockBaseModel)
    })
    return mocker

def test_is_pydantic_obj_with_pydantic_model(request, mocker_pydantic):
    logger.info(f'{request.node.name}()')

    # Create an instance of the mock BaseModel
    obj = MockBaseModel()

    # Call the function and assert the result
    result = _is_pydantic_obj(obj)
    assert result is True

def test_is_pydantic_obj_with_pydantic_v1_unavailable(request, mocker):
    logger.info(f'{request.node.name}()')

    # Mock the availability of pydantic v1
    mocker.patch('alexber.utils.thread_locals._is_available_pydantic_v1', False)
    mocker.patch('alexber.utils.thread_locals._is_available_pydantic_v2', True)

    # Mock the import of pydantic.v1 to raise ImportError
    mocker.patch.dict('sys.modules', {'pydantic.v1': None})
    mocker.patch.dict('sys.modules', {'pydantic': mocker.MagicMock(BaseModel=MockBaseModel)})

    # Create an instance of the mock BaseModel
    obj = MockBaseModel()

    # Call the function and assert the result
    result = _is_pydantic_obj(obj)
    assert result is True

def test_is_pydantic_obj_with_pydantic_v2_unavailable(request, mocker):
    logger.info(f'{request.node.name}()')

    # Mock the availability of pydantic v2
    mocker.patch('alexber.utils.thread_locals._is_available_pydantic_v2', False)
    mocker.patch('alexber.utils.thread_locals._is_available_pydantic_v1', True)

    # Mock the import of pydantic to raise ImportError
    mocker.patch.dict('sys.modules', {'pydantic': None})
    mocker.patch.dict('sys.modules', {'pydantic.v1': mocker.MagicMock(BaseModel=MockBaseModel)})

    # Create an instance of the mock BaseModel
    obj = MockBaseModel()

    # Call the function and assert the result
    result = _is_pydantic_obj(obj)
    assert result is True

def test_is_pydantic_obj_with_import_error(request, mocker):
    logger.info(f'{request.node.name}()')

    # Mock the import of pydantic to raise ImportError
    mocker.patch.dict('sys.modules', {'pydantic': None, 'pydantic.v1': None})

    # Create an instance of the mock BaseModel
    obj = MockBaseModel()

    # Call the function and assert the result
    result = _is_pydantic_obj(obj)
    assert result is False

def test_is_pydantic_obj_with_non_pydantic_object(request, mocker_pydantic):
    logger.info(f'{request.node.name}()')

    # Create an instance of a non-pydantic object
    class NonPydanticObject:
        pass

    obj = NonPydanticObject()

    # Call the function and assert the result
    result = _is_pydantic_obj(obj)
    assert result is False

def test_is_pydantic_obj_with_pydantic_object_mixin(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_pydantic_obj', return_value=True)
    obj = mocker.Mock()
    mixin = LockingPedanticObjMixin(obj=obj)
    assert mixin._is_pedantic_obj is True

def test_is_pydantic_obj_with_non_pydantic_object_mixin(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mocker.patch('alexber.utils.thread_locals._is_pydantic_obj', return_value=False)
    obj = mocker.Mock()
    mixin = LockingPedanticObjMixin(obj=obj)
    assert mixin._is_pedantic_obj is False

# Define a sample function
def sample_function(x, y):
    print("sample_function")
    return x + y

# Define a sample coroutine
async def sample_coroutine(x, y):
    await asyncio.sleep(1)
    return x * y

@pytest.mark.asyncio
async def test_exec_in_executor_with_function(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    # Run the sample function in an executor
    result = await exec_in_executor(None, sample_function, 3, 4)
    assert result == 7
    logger.info(f"Result from sample_function: {result}")

@pytest.mark.asyncio
async def test_exec_in_executor_with_coroutine(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    # Run the sample coroutine in an executor
    result = await exec_in_executor(None, sample_coroutine, 3, 4)
    assert result == 12
    logger.info(f"Result from sample_coroutine: {result}")

# Define the executor globally
EXECUTOR = ThreadPoolExecutor()

@pytest.fixture(scope="module")
def executor():
    yield EXECUTOR
    EXECUTOR.shutdown(wait=True)

def simplfified_blocking_io(x: int, y: int) -> int:
    #time.sleep(5)  # mimicking blocking I/O call
    return x + y


def blocking_io():
    #time.sleep(5)  # mimicking blocking I/O call

    async def helper_example_usage():
        #time.sleep(15)
        result = await sample_coroutine(2, 3)  # Example values for x and y
        return result

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(helper_example_usage())


@pytest.mark.asyncio
async def test_blocking_io(executor, request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    # Test the simplfified_blocking_io function to ensure it returns the correct result
    result = await exec_in_executor(executor, simplfified_blocking_io, 2, 3)
    assert result == 5, "The simplfified_blocking_io() function should return the sum of the inputs"

def some_legacy_function():
    fut = exec_in_executor_threading_future(EXECUTOR, blocking_io)
    result = fut.result()
    return result

@pytest.mark.asyncio
async def test_blocking_io_in_sync_context(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    result = some_legacy_function()
    assert result == 6, "The blocking_io() function should return the product of the inputs"

@pytest.mark.asyncio
async def test_blocking_io_in_async_context(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    result = await exec_in_executor(EXECUTOR, blocking_io)
    assert result == 6, "The blocking_io() function should return the product of the inputs"

@pytest.mark.asyncio
async def test_coroutine_in_async_context(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    result = await exec_in_executor(EXECUTOR, sample_coroutine, 2, 3)
    assert result == 6, "The sample_coroutine() function should return the sum of the inputs"

def another_legacy_function():
    fut = exec_in_executor_threading_future(EXECUTOR, sample_coroutine, 2, 3)
    result = fut.result()
    return result
@pytest.mark.asyncio
async def test_coroutine_in_sync_context(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    result = another_legacy_function()
    assert result == 6, "The sample_coroutine() function should return the sum of the inputs"

def worker_thread():
    print(f"Worker thread: {threading.current_thread().name} started.")
    # Execute task on the main thread's event loop
    loop = get_main_event_loop()
    future = asyncio.run_coroutine_threadsafe( sample_coroutine(2, 3),  loop=loop)  # Pass arguments to the coroutine
    # Wait for the result of the future
    result = future.result()
    print(f"Worker thread got result: {result}")
    return result


def worker_thread_switch_to_async():
    #time.sleep(5)  # mimicking blocking I/O call

    async def helper_example_usage():
        #time.sleep(15)
        loop = get_main_event_loop()
        threading_future = asyncio.run_coroutine_threadsafe( sample_coroutine(2, 3),  loop=loop)  # Pass arguments to the coroutine
        asyncio_future = asyncio.wrap_future(threading_future)
        result = await asyncio_future
        return result

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(helper_example_usage())

# Fixture to manage the event loop for the test
@pytest.fixture
def setup_event_loop():
    import alexber.utils.thread_locals as _thread_locals

    # Store the previous event loop
    prev_event_loop = _thread_locals._EVENT_LOOP

    with ThreadPoolExecutor(1) as executor:
        # Yield control to the test
        yield executor

    # Restore the previous event loop after the test
    _thread_locals._EVENT_LOOP = prev_event_loop

@pytest.mark.asyncio
async def test_run_coroutine_threadsafe_in_sync_context(executor, request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    # Set up the event loop in thread locals after the event loop is running
    import alexber.utils.thread_locals as _thread_locals
    _thread_locals._EVENT_LOOP = asyncio.get_running_loop()

    result = await exec_in_executor(executor, worker_thread)
    assert result == 6

@pytest.mark.asyncio
async def test_run_coroutine_threadsafe_switch_to_async_in_sync_context(executor, request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    # Set up the event loop in thread locals after the event loop is running
    import alexber.utils.thread_locals as _thread_locals
    _thread_locals._EVENT_LOOP = asyncio.get_running_loop()

    result = await exec_in_executor(executor, worker_thread_switch_to_async)
    assert result == 6


def worker_athread_switch_to_async():
    #time.sleep(5)  # mimicking blocking I/O call

    async def helper_example_usage():
        #time.sleep(15)
        loop = get_main_event_loop()
        threading_future = asyncio.run_coroutine_threadsafe( sample_coroutine(2, 3),  loop=loop)  # Pass arguments to the coroutine
        asyncio_future = asyncio.wrap_future(threading_future)
        result = await asyncio_future
        return result

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(helper_example_usage())
@pytest.mark.asyncio
async def test_arun_coroutine_threadsafe_switch_to_async_in_sync_context(executor, request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    # Set up the event loop in thread locals after the event loop is running
    import alexber.utils.thread_locals as _thread_locals
    _thread_locals._EVENT_LOOP = asyncio.get_running_loop()

    result = await exec_in_executor(executor, worker_athread_switch_to_async)
    logger.info('4')
    assert result == 6


if __name__ == "__main__":
    pytest.main([__file__])



import logging
import setuptools
from shutil import rmtree
import os
import sys

logger = logging.getLogger(__name__)

_setup_module = sys.modules['__main__']

_setup_dir = os.path.dirname(os.path.realpath(_setup_module.__file__)) if hasattr(_setup_module, '__file__') \
                                                                        else os.getcwd()
VERSION = getattr(_setup_module, 'VERSION', None)

def _my_rmtree(path, ignore_errors=False, onerror=None):
    try:
        rmtree(path, ignore_errors, onerror)
    except OSError:
        pass

#adapted from https://github.com/kennethreitz/setup.py/blob/master/setup.py
class UploadCommand(setuptools.Command):
    '''
    Support setup.py upload.
    UploadCommand is intented to be used only from setup.py
    It's builds Source and Wheel distribution.
    It's uploads the package to PyPI via Twine.
    It's pushes the git tags.
    '''

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f'\033[1m{s}\033[0m')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.status('Removing previous builds...')
        # rm -rf build *.egg-info dist
        _my_rmtree(os.path.join(_setup_dir, 'build'))
        _my_rmtree(os.path.join(_setup_dir, 'f{NAME}.egg-info'))
        _my_rmtree(os.path.join(_setup_dir, 'dist'))

        self.status('Building Source and Wheel distribution...')
        os.system(f'{sys.executable} setup.py sdist bdist_wheel')
        #os.system('python3 setup.py sdist bdist_wheel')

        self.status('Uploading the package to PyPI via Twine...')
        # python -m keyring set https://upload.pypi.org/legacy/ alex-ber
        os.system('twine upload dist/*')

        self.status('Pushing git tags...')
        os.system('git fetch')
        os.system('git commit -m "setup.py changed" setup.py')
        os.system('git push')
        if VERSION is not None:
            os.system(f'git tag v{VERSION}')
            os.system(f'git push origin v{VERSION}')
        #os.system(f'git tag -d v{VERSION}')
        #os.system(f'git push --delete origin v{VERSION}')
        #os.system('git push --tags')

        sys.exit()

class CleanCommand(setuptools.Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for dirpath, dirnames, filenames in os.walk('.'):
            for filename in filenames:
                if filename.endswith(('.pyc', '.pyo', '.pyd')):
                    file_path = os.path.join(dirpath, filename)
                    print(f'Removing {file_path}')
                    os.remove(file_path)
            for dirname in dirnames:
                if dirname == '__pycache__':
                    dir_path = os.path.join(dirpath, dirname)
                    print(f'Removing {dir_path}')
                    _my_rmtree(dir_path)
        _my_rmtree('build', ignore_errors=True)
        _my_rmtree('dist', ignore_errors=True)
        _my_rmtree('*.egg-info', ignore_errors=True)



import logging
import shlex
import os as _os
import tempfile
from pathlib import Path
import pytest
import alexber.utils.processinvokes as processinvokes
from alexber.utils.processinvokes import LogPipe, LogSubProcessCall

logger = logging.getLogger(__name__)
process_invokes_logger = None
_process_invokes_logger_log = None

@pytest.fixture
def mock_log(mocker):
    ret_mock = mocker.patch.object(process_invokes_logger, 'log', side_effect=_process_invokes_logger_log, autospec=True, spec_set=True)
    return ret_mock


@pytest.fixture
def mock_file(mocker):
    open_mock = mocker.patch('.'.join(['alexber.utils.processinvokes', 'open']), create=True)
    mock_close = mocker.MagicMock()
    mock_write = mocker.MagicMock()
    open_mock.return_value.close = mock_close
    open_mock.return_value.write = mock_write

    return open_mock


def _reset_processinvokes():
    processinvokes.default_log_name  = None
    processinvokes.default_log_level = None
    processinvokes.default_logpipe_cls = None

    executor = processinvokes.executor
    if executor is not None:
        executor.shutdown(wait=False)
        # see https://gist.github.com/clchiou/f2608cbe54403edb0b13
        #import concurrent.futures.thread
        #executor._threads.clear()
        #concurrent.futures.thread._threads_queues.clear()
    processinvokes.executor = None

@pytest.fixture
def processinvokesFixture(mocker):
    _reset_processinvokes()

    global process_invokes_logger
    process_invokes_logger = logging.getLogger('process_invoke_run')
    global _process_invokes_logger_log
    _process_invokes_logger_log = process_invokes_logger.log
    processinvokes.initConfig(**{'default_log_name': 'process_invoke_run'})
    yield None
    _reset_processinvokes()



def test_process_invokes(request, mocker, processinvokesFixture, mock_log):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_log_msg = "simulating run_sub_process"
    process_invoke_run = f"echo '{exp_log_msg}'"
    cmd = shlex.split(process_invoke_run)

    process_invoke_cwd = _os.getcwd()
    processinvokes.run_sub_process(*cmd, **{'kwargs':{'cwd':process_invoke_cwd}})

    pytest.assume(mock_log.call_count == 1)
    (_,logmsg), _ = mock_log.call_args
    pytest.assume(exp_log_msg == logmsg)

class MyLogPipe(LogPipe):
    pass

class MyLogSubProcessCall(LogSubProcessCall):
    pass

def test_init_config(request, mocker, processinvokesFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    logPipeClassName = '.'.join([__name__, MyLogPipe.__name__])
    logSubProcessCallClassName = '.'.join([__name__, MyLogSubProcessCall.__name__])

    processinvokes.initConfig(**{'default_log_name': 'process_invoke_run',
                                 'default_logpipe_cls': logPipeClassName,
                                 'default_log_subprocess_cls': logSubProcessCallClassName,
                                 })

    pytest.assume(processinvokes.default_logpipe_cls == MyLogPipe)
    pytest.assume(processinvokes.default_log_subprocess_cls == MyLogSubProcessCall)


def test_process_invokes_file_pipe1(request, mocker, processinvokesFixture, mock_file):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_log_msg = "simulating run_sub_process"
    process_invoke_run = f"echo '{exp_log_msg}'"
    cmd = shlex.split(process_invoke_run)

    process_invoke_cwd = _os.getcwd()
    filename = "my.log"
    mock_close = mocker.MagicMock()
    mock_write = mocker.MagicMock()
    mock_file.close = mock_close
    mock_file.write = mock_write
    processinvokes.initConfig(**{'default_log_name': 'process_invoke_run',
                                 'default_logpipe_cls': 'alexber.utils.processinvokes.FilePipe'
                                 })

    processinvokes.run_sub_process(*cmd, **{'cwd':process_invoke_cwd,
                                                      'logPipe': {
                                                          'kwargs' : {'fileName': filename}
                                                      }
                                                      })

    pytest.assume(mock_file.return_value.close.call_count == 1)
    pytest.assume(mock_file.return_value.write.call_count > 0)
    (logmsg,), _ = mock_file.return_value.write.call_args
    pytest.assume(f'{exp_log_msg}\n' == logmsg)

def test_process_invokes_file_pipe2(request, mocker, processinvokesFixture, mock_file):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_log_msg = "simulating run_sub_process"
    process_invoke_run = f"echo '{exp_log_msg}'"
    cmd = shlex.split(process_invoke_run)

    process_invoke_cwd = _os.getcwd()
    filename = "my.log"


    processinvokes.run_sub_process(*cmd, **{'cwd':process_invoke_cwd,
                                                  'logPipe': {
                                                      'cls': 'alexber.utils.processinvokes.FilePipe',
                                                      'kwargs' : {'fileName': filename}
                                                  }
                                                  })

    pytest.assume(mock_file.return_value.close.call_count == 1)
    pytest.assume(mock_file.return_value.write.call_count > 0)
    (logmsg,), _ = mock_file.return_value.write.call_args
    pytest.assume(f'{exp_log_msg}\n' == logmsg)



if __name__ == "__main__":
    pytest.main([__file__])



import logging
import concurrent.futures
import contextvars
import functools
import inspect
from concurrent.futures import Executor, Future
from contextvars import copy_context
from typing import Callable, Optional, TypeVar, Awaitable, Union
from threading import local
import asyncio
import threading
from collections import deque

# Define type variables for the function signature
T = TypeVar('T')

# Define a type that can be awaited and will return T, compatible with both asyncio.Future and concurrent.futures.Future
FutureType = Union[Awaitable[T], concurrent.futures.Future]

logger = logging.getLogger(__name__)


#should be initlialzied via initConfig() method, see below.
_EVENT_LOOP = None


# _GLOBAL_EXECUTOR is a global variable that holds the default executor for executing tasks
# when no specific executor is provided to the exec_in_executor() function.
#
# This allows for centralized management of task execution resources, ensuring that tasks are
# executed in a consistent and controlled manner across the application.
#
# The executor can be set during the initialization of the application or through configuration
# functions like initConfig(), providing flexibility in how tasks are managed and executed.
#
# If _GLOBAL_EXECUTOR is not set, the default asyncio executor will be used.
_GLOBAL_EXECUTOR = None

# Thread-local storage to hold the event loop for each thread
_event_loops_thread_locals = local()

# _CLOSE_SENTINEL is a unique object used as a marker to signal the worker to stop processing tasks.
# It is enqueued into the task queue to indicate that no more tasks will be added and the worker should exit.
# This approach allows for a clean shutdown of the worker loop without requiring additional control structures.
_CLOSE_SENTINEL = object()

def threadlocal_var(thread_locals, varname, factory, *args, **kwargs):
  v = getattr(thread_locals, varname, None)
  if v is None:
    v = factory(*args, **kwargs)
    setattr(thread_locals, varname, v)
  return v

def get_threadlocal_var(thread_locals, varname):
    v = threadlocal_var(thread_locals, varname, lambda : None)
    if v is None:
        raise ValueError(f"threadlocal's {varname} is not initilized")
    return v

def del_threadlocal_var(thread_locals, varname):
    try:
        delattr(thread_locals, varname)
    except AttributeError:
        pass


class RootMixin:
    """
    A mixin class that serves as the root of a delegation chain.

    The `RootMixin` class provides an initializer that stops the delegation chain,
    ensuring that no further delegation occurs.

    Methods:
    __init__(**kwargs): Initializes the mixin and stops the delegation chain.
    """
    def __init__(self, **kwargs):
        # The delegation chain stops here
        pass

def validate_param(param_value, param_name):
    """
     Validates that a required parameter is not None.

     param_value (any): The value of the parameter to be validated.
     param_name (str): The name of the parameter to be used in the error message.

     Raises:
     ValueError: If the parameter value is None, indicating that the required parameter is missing.
     """
    if param_value is None:
        raise ValueError(f"Expected {param_name} param not found")



class RLock:
    """
    A reentrant lock that supports both synchronous and asynchronous operations.
    The `RLock` class provides mechanisms to acquire and release locks in both synchronous
    and asynchronous contexts, ensuring proper synchronization and reentrancy.

    See https://alex-ber.medium.com/a6b9a9021be8 for more details.
    """

    def __init__(self):
        """
        Initializes the RLock instance with both synchronous and asynchronous locks.
        """
        self._sync_lock = threading.RLock()  # Synchronous reentrant lock
        self._async_lock = asyncio.Lock()  # Asynchronous lock

        self._sync_owner = None  # Owner of the synchronous lock
        self._async_owner = None  # Owner of the asynchronous lock

        self._sync_count = 0  # Reentrancy count for synchronous lock
        self._async_count = 0  # Reentrancy count for asynchronous lock

        self._sync_condition = threading.Condition(self._sync_lock)  # Condition variable for synchronization
        self._async_condition = asyncio.Condition()  # Asynchronous condition variable

        self._sync_waiting = deque()  # Queue for waiting synchronous threads
        self._async_waiting = deque()  # Queue for waiting asynchronous tasks

    def acquire(self):
        """
        Acquires the synchronous lock, blocking until it is available.
        Returns:
            bool: True if the lock was successfully acquired.
        """
        with self._sync_condition:
            current_thread = threading.current_thread()
            if self._sync_owner is current_thread:
                self._sync_count += 1
                return True  # Already acquired, no need to acquire again

            self._sync_waiting.append(current_thread)
            while self._sync_owner is not None or self._sync_waiting[0] != current_thread:
                self._sync_condition.wait()  # Wait until the lock is available

            self._sync_waiting.popleft()
            self._sync_owner = current_thread
            self._sync_count = 1
            return True  # Successfully acquired

    def release(self):
        """
        Releases the synchronous lock.
        Returns:
            bool: True if the lock was successfully released.
        Raises:
            RuntimeError: If the current thread does not own the lock.
        """
        with self._sync_condition:
            current_thread = threading.current_thread()
            if self._sync_owner is current_thread:
                self._sync_count -= 1
                if self._sync_count == 0:
                    self._sync_owner = None
                    self._sync_condition.notify_all()  # Notify all waiting threads
                return True  # Successfully released
            else:
                raise RuntimeError("Cannot release a lock that's not owned by the current thread")

    async def async_acquire(self):
        """
        Acquires the asynchronous lock, blocking until it is available.
        Returns:
            bool: True if the lock was successfully acquired.
        """
        async with self._async_condition:
            current_task = asyncio.current_task()
            if self._async_owner is current_task:
                self._async_count += 1
                return True  # Already acquired, no need to acquire again

            self._async_waiting.append(current_task)

            while self._async_owner is not None or self._async_waiting[0] != current_task:
                await self._async_condition.wait()  # Wait until the lock is available

            self._async_waiting.popleft()  # Remove the task from waiting queue once it acquires the lock
            self._async_owner = current_task
            self._async_count = 1
            return True  # Successfully acquired

    async def async_release(self):
        """
        Releases the asynchronous lock.
        Returns:
            bool: True if the lock was successfully released.
        Raises:
            RuntimeError: If the current task does not own the lock.
        """
        async with self._async_condition:
            current_task = asyncio.current_task()
            if self._async_owner is current_task:
                self._async_count -= 1
                if self._async_count == 0:
                    self._async_owner = None
                    self._async_condition.notify_all()  # Notify all waiting tasks
                return True  # Successfully released
            else:
                raise RuntimeError("Cannot release a lock that's not owned by the current task")

    def __enter__(self):
        """
        Enters the runtime context related to this object, acquiring the synchronous lock.
        """
        self.acquire()
        return self

    def __exit__(self, exc_type, exc, tb):
        """
        Exits the runtime context related to this object, releasing the synchronous lock.
        """
        self.release()

    async def __aenter__(self):
        """
        Enters the runtime context related to this object, acquiring the asynchronous lock.
        """
        await self.async_acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        Exits the runtime context related to this object, releasing the asynchronous lock.
        """
        await self.async_release()

class LockingIterableMixin(RootMixin):
    """
    A mixin class that provides locking for iterable objects.

    The `LockingIterableMixin` class ensures that iteration over the wrapped
    iterable object is thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the iterable object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the iterable object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        self._lock = kwargs.get('lock')
        validate_param(self._lock, 'lock')

    def __iter__(self):
        """
        Returns a locking iterator for the iterable object.

        Returns:
        LockingIterator: A locking iterator for the iterable object.
        """
        return LockingIterator(iter(self._obj), self._lock)

class LockingIterator:
    """
    An iterator that provides locking for thread-safe iteration.

    The `LockingIterator` class ensures that iteration over the wrapped
    iterator is thread-safe by using a provided lock.

    """
    def __init__(self, iterator, lock):
        """
        Initializes the iterator with the wrapped iterator and lock.

        Parameters:
        iterator (iterator): The wrapped iterator.
        lock (RLock): The lock to be used for synchronization.
        """
        self._iterator = iterator
        self._lock = lock

    def __iter__(self):
        """
        Returns the iterator itself.

        Returns:
        LockingIterator: The iterator itself.
        """
        return self

    def __next__(self):
        """
        Returns the next item from the iterator, acquiring the lock.

        Returns:
        any: The next item from the iterator.

        Raises:
        StopIteration: If the iterator is exhausted.
        """
        with self._lock:
            return next(self._iterator)

class LockingAsyncIterableMixin(RootMixin):
    """
    A mixin class that provides locking for asynchronous iterable objects.

    The `LockingAsyncIterableMixin` class ensures that iteration over the wrapped
    asynchronous iterable object is thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the asynchronous iterable object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the asynchronous iterable object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        self._lock = kwargs.get('lock')
        validate_param(self._lock, 'lock')

    def __aiter__(self):
        """
        Returns a locking asynchronous iterator for the iterable object.

        Returns:
        LockingAsyncIterator: A locking asynchronous iterator for the iterable object.
        """
        return LockingAsyncIterator(self._obj, self._lock)


class LockingAsyncIterator:
    """
    An asynchronous iterator that provides locking for thread-safe iteration.

    The `LockingAsyncIterator` class ensures that iteration over the wrapped
    asynchronous iterator is thread-safe by using a provided lock.

    """
    def __init__(self, async_iterator, lock):
        """
        Initializes the iterator with the wrapped asynchronous iterator and lock.

        Parameters:
        async_iterator (async iterator): The wrapped asynchronous iterator.
        lock (RLock): The lock to be used for synchronization.
        """
        self._async_iterator = async_iterator
        self._lock = lock

    def __aiter__(self):
        """
        Returns the iterator itself.

        Returns:
        LockingAsyncIterator: The iterator itself.
        """
        return self

    async def __anext__(self):
        """
        Returns the next item from the iterator, acquiring the lock.

        Returns:
        any: The next item from the iterator.

        Raises:
        StopAsyncIteration: If the iterator is exhausted.
        """
        async with self._lock:
            return await self._async_iterator.__anext__()

class LockingPedanticObjMixin(RootMixin):
    """
    A mixin class that provides locking for Pydantic objects.

    The `LockingPedanticObjMixin` class ensures that access to the wrapped
    Pydantic object is thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the Pydantic object.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the Pydantic object.
        """
        super().__init__(**kwargs)
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        self._is_pedantic_obj = _is_pydantic_obj(self._obj)

class LockingGetItemMixin(RootMixin):
    """Mixin to provide thread-safe access and modification for collection-like objects.

    This mixin ensures that the `__getitem__` and `__setitem__` methods are thread-safe by using a lock.
    It is intended to be used with objects that support item access and modification, such as lists or dictionaries.

    Attributes:
        _obj: The object being proxied, which supports item access and modification.
        _lock: A threading lock to ensure thread-safe operations.
    """

    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)

    def __getitem__(self, key):
        """Retrieve an item from the collection in a thread-safe manner.

        Parameters:
            key: The key or index of the item to retrieve.

        Returns:
            The item associated with the given key.
        """
        @functools.wraps(self._obj.__getitem__)
        def call(key):
            with self._lock:
                return self._obj[key]
        return call(key)

    def __setitem__(self, key, value):
        """Set an item in the collection in a thread-safe manner.

        Parameters:
            key: The key or index of the item to set.
            value: The value to associate with the given key.

        Returns:
            The value that was set.
        """
        @functools.wraps(self._obj.__setitem__)
        def call(key, value):
            with self._lock:
                self._obj[key] = value
            return value
        return call(key, value)


class LockingSetItemMixin(RootMixin):
    """Mixin to provide thread-safe modification of items in a collection.

    This mixin ensures that the `__setitem__` method is thread-safe by using a lock.
    It is intended to be used with objects that support item modification, such as lists or dictionaries.

    Attributes:
        _obj: The object being proxied, which supports item modification.
        _lock: A threading lock to ensure thread-safe modification.
    """

    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)

    def __setitem__(self, key, value):
        """Set an item in the collection in a thread-safe manner.

        Parameters:
            key: The key or index of the item to set.
            value: The value to associate with the given key.

        Returns:
            The value that was set.
        """
        @functools.wraps(self._obj.__setitem__)
        def call(key, value):
            with self._lock:
                self._obj[key] = value
            return value
        return call(key, value)

class SyncContextManagerMixin:
    """
    Mixin to add synchronous context management to a class.
    """

    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)
    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()

class AsyncContextManagerMixin:
    """
    Mixin to add asynchronous context management to a class.
    """

    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)

    async def __aenter__(self):
        await self._lock.async_acquire()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._lock.async_release()

class LockingAccessMixin(LockingPedanticObjMixin):
    """
    A mixin class that provides locking for attribute access.

    The `LockingAccessMixin` class ensures that access to the attributes of the
    wrapped object is thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        self._lock = kwargs.get('lock')
        validate_param(self._lock, 'lock')

    def __getattr__(self, name):
        """
        Returns the attribute, acquiring the lock if necessary.

        Parameters:
        name (str): The name of the attribute.

        Returns:
        any: The attribute value.
        """
        attr = getattr(self._obj, name)
        if inspect.isroutine(attr):
            if self._is_pedantic_obj and name == '_copy_and_set_values':
                # special case for Pydantic
                @functools.wraps(attr)
                def synchronized_method(*args, **kwargs):
                    with self._lock:
                        attr(*args, **kwargs)
                    return self
                return synchronized_method
            elif inspect.iscoroutinefunction(attr):
                @functools.wraps(attr)
                async def asynchronized_method(*args, **kwargs):
                    async with self._lock:
                        return await attr(*args, **kwargs)
                return asynchronized_method
            else:
                @functools.wraps(attr)
                def synchronized_method(*args, **kwargs):
                    with self._lock:
                        return attr(*args, **kwargs)
                return synchronized_method
        elif hasattr(attr, '__get__') or hasattr(attr, '__set__') or hasattr(attr, '__delete__'):
            # Handle property or descriptor
            if hasattr(attr, '__get__'):
                return attr.__get__(self._obj, type(self._obj))
            return attr
        else:
            return attr

class LockingCallableMixin(RootMixin):
    """
    A mixin class that provides locking for callable objects.

    The `LockingCallableMixin` class ensures that calls to the wrapped
    callable object are thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the callable object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the callable object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        self._lock = kwargs.get('lock')
        validate_param(self._lock, 'lock')

    def __call__(self, *args, **kwargs):
        """
        Calls the wrapped callable object, acquiring the lock.

        Parameters:
        *args: Positional arguments for the callable object.
        **kwargs: Keyword arguments for the callable object.

        """
        if inspect.iscoroutinefunction(self._obj):
            @functools.wraps(self._obj)
            async def acall(*args, **kwargs):
                async with self._lock:
                    return await self._obj(*args, **kwargs)
            return acall(*args, **kwargs)
        else:
            @functools.wraps(self._obj)
            def call(*args, **kwargs):
                with self._lock:
                    return self._obj(*args, **kwargs)
            return call(*args, **kwargs)

class LockingDefaultLockMixin(RootMixin):
    """
    A mixin class that provides a default lock if none is provided.

    The `LockingDefaultLockMixin` class ensures that a default lock is used
    if no lock is provided during initialization.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        lock = kwargs.get("lock", None)
        if not lock:
            lock = RLock()
        kwargs['lock'] = lock
        self._lock = lock
        super().__init__(**kwargs)

def _coerce_base_language_model(proxy):
    """
    Coerces the proxy object to be recognized as a BaseLanguageModel.

    Parameters:
    proxy (LockingProxy): The proxy object to be coerced.
    """
    if not _is_available_base_language_model:
        return
    if isinstance(proxy._obj, BaseLanguageModel):
        BaseLanguageModel.register(type(proxy))

class LockingBaseLanguageModelMixin(RootMixin):
    """
    A mixin class that provides locking for BaseLanguageModel objects.

    The `LockingBaseLanguageModelMixin` class ensures that access to the wrapped
    BaseLanguageModel object is thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the BaseLanguageModel object.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the BaseLanguageModel object.
        """
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        _coerce_base_language_model(self)
        super().__init__(**kwargs)

class LockingDefaultAndBaseLanguageModelMixin(LockingDefaultLockMixin, LockingBaseLanguageModelMixin):
    """
    A mixin class that combines default lock and BaseLanguageModel locking.

    The `LockingDefaultAndBaseLanguageModelMixin` class ensures that a default lock is used
    if no lock is provided and that access to the wrapped BaseLanguageModel object is thread-safe.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)


class LockingBaseLanguageModelMixin(RootMixin):
    """
    A mixin class that provides locking for BaseLanguageModel objects.

    The `LockingBaseLanguageModelMixin` class ensures that access to the wrapped
    BaseLanguageModel object is thread-safe by using a provided lock.

    """
    def __init__(self, **kwargs):
        """
        Initializes the mixin with the BaseLanguageModel object.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the BaseLanguageModel object.
        """
        self._obj = kwargs.get('obj')
        validate_param(self._obj, 'obj')
        _coerce_base_language_model(self)
        super().__init__(**kwargs)



# Flags to check the availability of Pydantic and BaseLanguageModel
try:
    from langchain_core.language_models import BaseLanguageModel

    _is_available_base_language_model = True
except ImportError:
    BaseLanguageModel = None
    _is_available_base_language_model = False

try:
    from pydantic import BaseModel

    _is_available_pydantic_v2 = True
except ImportError:
    _is_available_pydantic_v2 = False

try:
    from pydantic.v1 import BaseModel

    _is_available_pydantic_v1 = True
except ImportError:
    _is_available_pydantic_v1 = False


def _is_pydantic_v1_obj(obj):
    """
    Checks if the object is a Pydantic v1 object.

    Parameters:
    obj (any): The object to be checked.

    Returns:
    bool: True if the object is a Pydantic v1 object, False otherwise.
    """
    if not _is_available_pydantic_v1:
        return False
    ret = None
    try:
        from pydantic.v1 import BaseModel as BaseModelv1
        ret = isinstance(obj, BaseModelv1)
    except ImportError:
        ret = False
    return ret


def _is_pydantic_v2_obj(obj):
    """
    Checks if the object is a Pydantic v2 object.

    Parameters:
    obj (any): The object to be checked.

    Returns:
    bool: True if the object is a Pydantic v2 object, False otherwise.
    """
    if not _is_available_pydantic_v2:
        return False
    ret = None
    try:
        from pydantic import BaseModel as BaseModelv2
        ret = isinstance(obj, BaseModelv2)
    except ImportError:
        ret = False
    return ret

def _is_pydantic_obj(obj):
    """
    Checks if the given object is an instance of either Pydantic v1 or v2 models.


    Args:
        obj (Any): The object to be checked.

    Returns:
        bool: True if the object is an instance of either Pydantic v1 or v2 models, False otherwise.
    """

    ret = _is_pydantic_v1_obj(obj) or _is_pydantic_v2_obj(obj)
    return ret

class LockingProxy(LockingDefaultAndBaseLanguageModelMixin,
                   LockingIterableMixin,
                   LockingAsyncIterableMixin,
                   LockingAccessMixin,
                   LockingCallableMixin,
                   LockingGetItemMixin,
                   LockingSetItemMixin,
                   SyncContextManagerMixin,
                   AsyncContextManagerMixin):
    """
    A proxy class that combines multiple locking mixins.

    The `LockingProxy` class ensures that access to the wrapped object is thread-safe
    by using a provided lock. It supports iterable, async iterable, attribute access,
    and callable objects, and thread-safe item access and modification
    and context management, async context management.

    See https://alex-ber.medium.com/7a7a14021427 for more details.

    """
    def __init__(self, **kwargs):
        """
        Initializes the proxy with the object and lock.

        Parameters:
        **kwargs: Arbitrary keyword arguments, including 'obj' for the object
                  and 'lock' for the lock.
        """
        super().__init__(**kwargs)


def is_running_in_main_thread():
    """
    Checks if the current thread is the main thread.

    Returns:
        bool: True if the current thread is the main thread, False otherwise.
    """
    logger.info("is_running_in_main_thread()")

    ret = threading.current_thread() is threading.main_thread()
    logger.info(f"Going to return {ret}")
    return ret


def _execute_async_in_sync(afunc_call):
    """
    Executes an asynchronous function call in a synchronous context.

    Args:
        afunc_call: The asynchronous function call to be executed.

    Returns:
        The result of the asynchronous function call.

    Raises:
        RuntimeError: If the function is called from the main thread while an event loop is running.
    """
    logger.info("_execute_async_in_sync()")

    has_running_loop = False

    try:
        loop = asyncio.get_running_loop()
        has_running_loop = True
    except RuntimeError:
        logger.info("Resetting EVENT_LOOP")
        loop = _EVENT_LOOP
        asyncio.set_event_loop(loop)

    if has_running_loop:
        logger.info('The event loop is running.')
        if is_running_in_main_thread():
            s = """
            If we have a running event loop, we should be on a non-MainThread; otherwise, we will block the event loop. 
            The easiest way to achieve this is to use asyncio.to_thread() before making the call. 
            In any case, you should execute your synchronous function on a separate thread, so any blocking will occur 
            on that thread and not on the main thread that runs the event loop.			
            """

            raise RuntimeError(s)

        logger.info("But we're running on a non-MainThread, so we're good.")
        logger.info("Going to submit a coroutine to a given event loop from a different thread and execute it in a thread-safe way.")
        logger.info("Going to block current thread and wait for result")

        future = concurrent.futures.Future()

        # Define a callback to set the result of the concurrent.futures.Future
        def on_complete(task):
            if task.exception():
                future.set_exception(task.exception())
            else:
                future.set_result(task.result())

        # Schedule the coroutine and add the callback
        task = asyncio.run_coroutine_threadsafe(afunc_call(), loop)
        task.add_done_callback(lambda t: on_complete(t))
        result = future.result()

    else:
        logger.info("The event loop is NOT running.")
        logger.info("Submitting a coroutine to the event loop from a different thread in a thread-safe manner.")
        logger.info("This will start the event loop.")
        logger.info("The current thread will continue to run.")

        result = asyncio.run_coroutine_threadsafe(afunc_call(), loop).result()

    return result

def lift_to_async(afunc, /, *args, **kwargs):
    """
    Executes an asynchronous function in a synchronous context preserving ContextVar's context.

    Args:
        afunc: The asynchronous function to be executed.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The result of the asynchronous function call.
    """
    logger.info("lift_to_async()")

    #see https://github.com/alex-ber/AlexBerUtils/issues/14
    def wrapper():
        try:
            return afunc(*args, **kwargs)
        except StopAsyncIteration as exc:
            # StopIteration can't be set on an asyncio.Future
            # it raises a TypeError and leaves the Future pending forever
            # so we need to convert it to a RuntimeError
            raise RuntimeError("Async generator exhausted unexpectedly") from exc

    ctx = contextvars.copy_context()
    afunc_call = functools.partial(ctx.run, wrapper)
    result = _execute_async_in_sync(afunc_call)
    return result


def ensure_thread_event_loop():
    """
    Initializes an event loop for the current thread if it does not already exist.

    This function first checks if the current thread has an event loop stored in thread-local storage.
    If not, it attempts to retrieve the current event loop. If no event loop is present, it creates a new one
    and sets it as the current event loop for the thread. The event loop is then also stored in thread-local storage.
    """
    # Check if the current thread already has an event loop in thread-local storage
    if not hasattr(_event_loops_thread_locals, 'loop'):
        try:
            # Try to get the current event loop
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop is present, create a new one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Store the event loop in thread-local storage
        _event_loops_thread_locals.loop = loop

def _run_coroutine_in_thread(coro):
    """
    Runs a coroutine in the event loop of the current thread.

    This function ensures that the current thread has an event loop initialized. It then runs the given coroutine
    until it is complete using the thread's event loop.

    Args:
        coro: The coroutine to be executed.

    Returns:
        The result of the coroutine execution.
    """
    # Ensure the thread has an event loop
    ensure_thread_event_loop()
    loop = _event_loops_thread_locals.loop
    return loop.run_until_complete(coro)


def exec_in_executor(executor: Optional[Executor], func: Callable[..., T], *args, **kwargs) -> asyncio.Future:
    """
    Execute a function or coroutine within a given executor while preserving `ContextVars`, ensuring that context is maintained across asynchronous boundaries.


    The executor is resolved in the following order:
    1. If the `executor` parameter is provided, it is used.
    2. If an executor was passed via `initConfig()`, it is used.
    3. If neither is set, `None` is used, which means the default asyncio executor will be used.

    Note: as a side fact, threads in the executor may have an event loop attached. This allows for the execution of asynchronous tasks within those threads.

    Args:
        executor (Optional[Executor]): The executor to run the function or coroutine. If None, the default asyncio executor is used.
        func (Callable[..., T]): The function or coroutine to execute.
        *args: Positional arguments to pass to the function or coroutine.
        **kwargs: Keyword arguments to pass to the function or coroutine.

    Returns:
        asyncio.Future: A future representing the execution of the function or coroutine.
    """
    loop = asyncio.get_running_loop()
    resolved_executor = executor if executor is not None else _GLOBAL_EXECUTOR



    if asyncio.iscoroutinefunction(func):

        # Wrap the coroutine function to preserve metadata
        @functools.wraps(func)
        async def wrapped_coro(*args, **kwargs):
            # Run the coroutine in the thread's event loop
            return await ctx.run(func, *args, **kwargs)

        coro = wrapped_coro(*args, **kwargs)
        return loop.run_in_executor(resolved_executor, _run_coroutine_in_thread, coro)
    else:
        # Copy the current context
        ctx = copy_context()

        # Wrap the function or coroutine call with the context
        func_or_coro_call = functools.partial(ctx.run, func, *args, **kwargs)

        @functools.wraps(func)
        def wrapper() -> T:
            ensure_thread_event_loop()
            try:
                return func_or_coro_call()
            except StopIteration as exc:
                # StopIteration can't be set on an asyncio.Future
                # it raises a TypeError and leaves the Future pending forever
                # so we need to convert it to a RuntimeError
                raise RuntimeError from exc

        # If func is a regular function, run it in an executor guarded against StopIteration
        return loop.run_in_executor(resolved_executor, wrapper)

def exec_in_executor_threading_future(executor: Optional[Executor], func: Callable[..., T], *args, **kwargs) -> Future:
    """
    Execute a function or coroutine within a given executor and return a threading.Future.

    This function executes a function or coroutine while preserving `ContextVars`, ensuring that context is maintained across asynchronous boundaries. It provides a threading.Future to handle the result or exception of the task execution.

    The executor is resolved in the following order:
    1. If the `executor` parameter is provided, it is used.
    2. If an executor was passed via `initConfig()`, it is used.
    3. If neither is set, `None` is used, which means the default asyncio executor will be used.

    Note: as a side fact, threads in the executor may have an event loop attached. This allows for the execution of asynchronous tasks within those threads.

    Args:
        executor (Optional[Executor]): The executor to run the function or coroutine. If None, the default asyncio executor is used.
        func (Callable[..., T]): The function or coroutine to execute.
        *args: Positional arguments to pass to the function or coroutine.
        **kwargs: Keyword arguments to pass to the function or coroutine.

    Returns:
        threading.Future: A future representing the execution of the function or coroutine.
    """

    future = concurrent.futures.Future()

    if asyncio.iscoroutinefunction(func):
        @functools.wraps(func)
        async def wrapper(fut):
            coro = func(*args, **kwargs)
            result = await coro
            fut.set_result(result)
    else:
        @functools.wraps(func)
        def wrapper(fut):
            result = func(*args, **kwargs)
            fut.set_result(result)

    exec_in_executor(executor, wrapper, future)
    return future


def chain_future_results(source_future: FutureType, target_future: FutureType):
    """
    Transfers the result or exception from one future to another.

    This function is called when the source_future is completed. It retrieves the result or exception
    from the source_future and sets it on the target_future, ensuring that the outcome of the task execution
    is properly propagated. This function is generic and can be used with any types of futures.

    To use this function, add it as a callback to the source_future:

    # Add the chain_future_results function as a callback to the source_future
    source_future.add_done_callback(lambda fut: chain_future_results(fut, target_future))

    Args:
        source_future: The future from which to retrieve the result or exception.
        target_future: The future on which to set the result or exception.
    """
    try:
        result = source_future.result()
        target_future.set_result(result)
    except Exception as e:
        target_future.set_exception(e)

def get_main_event_loop():
    """
    Returns the event loop that was set via initConf(). This function should be called from the MainThread
    during the application startup. If such a call was not made, None will be returned.

    Returns:
        asyncio.AbstractEventLoop or None: The main event loop if set, otherwise None.
    """
    return _EVENT_LOOP

class AsyncExecutionQueue(RootMixin):
    """
    A class representing an asynchronous task queue that manages task execution using a specified executor.

    This class provides a context manager interface to start and stop a worker that processes tasks from the queue.
    Tasks are executed asynchronously, and the queue can be closed gracefully.

    The executor is resolved in the following order:
    1. If the `executor` parameter is provided, it is used.
    2. If an executor was passed via `initConfig()`, it is used.
    3. If neither is set, `None` is used, which means the default asyncio executor will be used.

    Note: as a side fact, threads in the executor may have an event loop attached. This allows for the execution of asynchronous tasks within those threads.


    Attributes:
        queue (asyncio.Queue): The queue that holds tasks to be executed.
        executor (Executor): The executor to run tasks.

    Methods:
        worker(): Continuously processes tasks from the queue until the `aclose()` method is called.
        aadd_task(func, *args, **kwargs): Asynchronously adds a task to the queue for execution and returns a future.
        aclose(): Asynchronously closes the queue and waits for the worker to finish processing.
    """

    def __init__(self, **kwargs):
        """
        Initializes the TaskQueue with a specified queue and executor.

        Args:
            **kwargs: Optional keyword arguments to configure the queue and executor.
                - queue (asyncio.Queue, optional): A custom queue to use. If not provided, a new asyncio.Queue is created.
                - executor (Executor): The executor to run tasks. This parameter is required.

        Execute a function or coroutine within a given executor while preserving `ContextVars`, ensuring that context is maintained across asynchronous boundaries.
        """
        self.queue = kwargs.pop("queue", None)
        if not self.queue:
            self.queue = asyncio.Queue()
        #if None, default asyncio ThreadPoolExecutor will be used.
        self.executor = kwargs.pop("executor", None)
        self._worker_task = None
        super().__init__(**kwargs)

    async def __aenter__(self):
        """
        Starts the worker when entering the context.
        """
        self._worker_task = asyncio.create_task(self.worker())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Closes the queue and stops the worker when exiting the context.
        """
        await self.aclose()

    async def worker(self):
        """
        Continuously processes tasks from the queue until the `aclose()` method is called.

        This method continuously fetches tasks from the queue and executes them asynchronously using the specified executor.
        Execute a function or coroutine within a given executor while preserving `ContextVars`, ensuring that context is maintained across asynchronous boundaries.
        """
        while True:
            task, task_future = await self.queue.get()
            try:
                if task is _CLOSE_SENTINEL:
                    return  # Exit the worker loop
                func, args, kwargs = task
                result_future = exec_in_executor(self.executor, func, *args, **kwargs)
                result_future.add_done_callback(lambda fut: chain_future_results(fut, task_future))
            finally:
                # Mark the task as done, regardless of what the task was
                self.queue.task_done()


    async def aadd_task(self, func, /, *args, **kwargs):
        """
        Asynchronously adds a task to the queue for execution and returns a future.

        Args:
            func (Callable): The function to be executed, which can be synchronous or asynchronous.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.

        Execute a function or coroutine within a given executor while preserving `ContextVars`, ensuring that context is maintained across asynchronous boundaries.

        Returns:
            asyncio.Future: A future representing the execution of the function or coroutine.
        """
        future = asyncio.get_running_loop().create_future()
        await self.queue.put(((func, args, kwargs), future))
        return future

    def add_task(self, executor, func, /, *args, **kwargs):
        """
        Adds a task to the queue for asynchronous execution and returns a future.

        The executor is resolved in the following order:
        1. If the `executor` parameter is provided, it is used.
        2. If an executor was passed via `initConfig()`, it is used.
        3. If neither is set, `None` is used, which means the default asyncio executor will be used.


        Args:
            executor (Executor): The executor to run the asynchronous task.
            func (Callable[..., Any]): The function to be executed, which can be synchronous or asynchronous.
            *args (Any): Positional arguments to pass to the function.
            **kwargs (Any): Keyword arguments to pass to the function.

        Execute a function or coroutine within a given executor while preserving `ContextVars`, ensuring that context is maintained across asynchronous boundaries.

        Returns:
            threading.Future: A future representing the execution of the function or coroutine.
        """
        fut = exec_in_executor_threading_future(executor, self.aadd_task, func, *args, **kwargs)
        return fut


    async def aclose(self):
        """
        Asynchronously closes the queue and waits for the worker to finish processing.

        This method signals the worker to stop processing tasks and waits for the worker task to complete.
        """
        await self.queue.put((_CLOSE_SENTINEL, None))
        if self._worker_task:
            await self._worker_task




def initConfig(**kwargs):
    """
    Initializes the configuration required for using the lift_to_async(), exec_in_executor() and get_event_loop() methods.

    This function is intended to be called from the MainThread.
    It can be called with empty parameters.
    It should be called with running event loop.

    Args:
        **kwargs: Optional keyword arguments to configure the initialization.

    Returns:
        None
    """
    _loop = asyncio.get_running_loop()  #raises exception, if there is no running event loop
    global _EVENT_LOOP
    _EVENT_LOOP = _loop

    global _GLOBAL_EXECUTOR
    # Set the global executor if provided
    _GLOBAL_EXECUTOR = kwargs.get('executor', None)




import warnings
import functools

# Define a custom warning for optional NumPy support
class OptionalNumpyWarning(Warning):
    """Custom warning to indicate that NumPy is not available and a fallback to standard Python is used."""



def deprecated(version=None, reason=None, for_removal=False, warning_class=DeprecationWarning):
    """
    A decorator to mark functions as deprecated.

    This decorator can be used to indicate that a function is deprecated and may be removed in future versions.
    It issues a warning when the decorated function is called.

    Args:
        version (str, optional): The version since which the function is considered deprecated.
        reason (str, optional): The reason why the function is deprecated.
        for_removal (bool, optional): If True, indicates that the function is marked for removal.
        warning_class (Warning, optional): The class of warning to be issued. Defaults to DeprecationWarning.

    Returns:
        function: The decorated function that issues a deprecation warning when called.
    """
    def decorator(func):
        message = f"{func.__name__} is deprecated"
        if version:
            message += f" since version {version}"
        if reason:
            message += f": {reason}"
        if for_removal:
            message += " and is marked for removal."
        else:
            message += " and will be removed in a future version."

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(message, warning_class, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator



import logging
import random as _random
import math
import warnings
from typing import Union, Optional
from .warning import OptionalNumpyWarning

# Initialize a logger for this module
logger = logging.getLogger(__name__)

class SamplingError(Exception):
    """Custom exception raised when sampling fails after maximum retries.

    Attributes:
        distribution (str): The distribution from which sampling was attempted.
        retries (int): The number of retries attempted before failure.
        lower_bound (Optional[float]): The lower bound for the sampled value.
        upper_bound (Optional[float]): The upper bound for the sampled value.
    """
    def __init__(self, message, distribution, retries, lower_bound, upper_bound):
        """
        Initialize the SamplingError with a message, distribution, retries, and bounds.

        :param message: The error message.
        :param distribution: The distribution from which sampling was attempted.
        :param retries: The number of retries attempted before failure.
        :param lower_bound: The lower bound for the sampled value.
        :param upper_bound: The upper bound for the sampled value.
        """
        super().__init__(message)
        self.distribution = distribution
        self.retries = retries
        self.lower_bound = lower_bound if lower_bound != -math.inf else None
        self.upper_bound = upper_bound if upper_bound != math.inf else None

    def __str__(self):
        """Return a user-friendly string representation of the error."""
        bounds_info = f"Lower bound: {self.lower_bound}, Upper bound: {self.upper_bound}"
        return (f"{self.args[0]} (Distribution: {self.distribution}, "
                f"Retries: {self.retries}, {bounds_info})")

    def __repr__(self):
        """Return a detailed string representation of the error for debugging."""
        bounds_info = f"lower_bound={self.lower_bound!r}, upper_bound={self.upper_bound!r}"
        return (f"SamplingError(message={self.args[0]!r}, distribution={self.distribution!r}, "
                f"retries={self.retries!r}, {bounds_info})")


# Try to import NumPy
try:
    import numpy as np
    USE_NUMPY = True
except ImportError:
    warning_message = (
        "NumPy module wasn't found. Falling back to standard Python. "
        "Using NumPy may lead to a performance boost. "
        "You can install it by running 'python -m pip install alex-ber-utils[numpy]'."
    )
    warnings.warn(warning_message, OptionalNumpyWarning)
    USE_NUMPY = False

class BaseSampler:
    """
    Base class for sampling from various statistical distributions with configurable parameters.

    Supported Distributions:
        - 'lognormvariate': Log-normal distribution.
        - 'normalvariate': Normal distribution.
        - 'expovariate': Exponential distribution.
        - 'vonmisesvariate': Von Mises distribution.
        - 'gammavariate': Gamma distribution.
        - 'gauss': Gaussian distribution.
        - 'betavariate': Beta distribution.
        - 'paretovariate': Pareto distribution.
        - 'weibullvariate': Weibull distribution.

    Attributes:
        distribution (str): The distribution to sample from.
        shape (Union[float, np.float32, np.float64]): Shape parameter for the distribution, controlling the spread and skewness.
                       For log-normal, it represents sigma of the underlying normal distribution.
        scale (Union[float, np.float32, np.float64]): Scale parameter for the distribution, shifting the distribution and determining its median.
                       For log-normal, it represents exp(mu) of the underlying normal distribution.
                       For exponential, it is used directly as the mean of the distribution.
        lower_bound (Optional[Union[float, np.float32, np.float64]]): Lower bound for the sampled value. Default is None (interpreted as unbounded).
        upper_bound (Optional[Union[float, np.float32, np.float64]]): Upper bound for the sampled value. Default is None (interpreted as unbounded).
        max_retries (int): Maximum number of attempts to sample a valid value. Default is 1000.
    """

    # Class-level attribute for supported distributions
    supported_distributions = {
        'lognormvariate', 'normalvariate', 'expovariate', 'vonmisesvariate',
        'gammavariate', 'gauss', 'betavariate', 'paretovariate', 'weibullvariate'
    }

    def __init__(self, **kwargs):
        """
        Initialize the BaseSampler with required and optional parameters.

        :param kwargs: Keyword arguments for initialization.
        """
        logger.info("__init__()")

        self.distribution = kwargs.get('distribution', None)
        self.shape = kwargs.get('shape', None)
        self.scale = kwargs.get('scale', None)
        self.lower_bound = kwargs.get('lower_bound', -math.inf)
        self.upper_bound = kwargs.get('upper_bound', math.inf)
        self.max_retries = kwargs.get('max_retries', 1000)

        self.validate_distribution()
        self.validate_bounds()

    def validate_distribution(self):
        """
        Validate that the specified distribution is supported.
        """
        if self.distribution not in self.supported_distributions:
            raise ValueError(f"Unsupported distribution: {self.distribution}")

    def validate_bounds(self):
        """
        Validate that the lower bound is less than the upper bound.
        """
        if not (self.lower_bound < self.upper_bound):
            raise ValueError("lower_bound must be less than upper_bound")

    def validate_random_parameters(self, seed, instance):
        """
        Validate that only one of random_seed or random_state/random_instance is provided.

        :param seed: The random seed.
        :param instance: The random state or instance.
        """
        if seed is not None and instance is not None:
            raise ValueError("Specify only one of random_seed or random_state/random_instance")

    def get_sample(self) -> Union[float, 'np.float32', 'np.float64']:
        """
        Get a sample from the specified distribution.

        :return: A sample from the specified distribution within the specified bounds.
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

if USE_NUMPY:
    class Sampler(BaseSampler):
        """
        A class to sample from various statistical distributions using NumPy.
        """

        def __init__(self, **kwargs):
            """
            Initialize the Sampler with NumPy-specific parameters.

            :param kwargs: Keyword arguments for initialization.
            """
            random_seed = kwargs.pop('random_seed', None)
            random_state = kwargs.pop('random_state', None)
            self.validate_random_parameters(random_seed, random_state)

            super().__init__(**kwargs)

            if random_state is not None:
                self.random_state = random_state
            else:
                self.random_state = np.random.RandomState(random_seed)

        def get_sample(self) -> Union[float, 'np.float32', 'np.float64']:
            """
            Get a sample from the specified distribution using NumPy.

            :return: A sample from the specified distribution within the specified bounds.
            """
            logger.info("get_sample()")

            distribution_methods = {
                'lognormvariate': lambda: self.random_state.lognormal(math.log(self.scale), self.shape),
                'normalvariate': lambda: self.random_state.normal(self.scale, self.shape),
                'expovariate': lambda: self.random_state.exponential(self.scale),
                'vonmisesvariate': lambda: self.random_state.vonmises(self.scale, self.shape),
                'gammavariate': lambda: self.random_state.gamma(self.shape, self.scale),
                'gauss': lambda: self.random_state.normal(self.scale, self.shape),
                'betavariate': lambda: self.random_state.beta(self.shape, self.scale),
                'paretovariate': lambda: self.random_state.pareto(self.shape),
                'weibullvariate': lambda: self.random_state.weibull(self.shape) * self.scale
            }

            for _ in range(self.max_retries):
                sampled_value = distribution_methods[self.distribution]()
                if self.lower_bound <= sampled_value <= self.upper_bound:
                    return sampled_value
            raise SamplingError(
                "Failed to sample a valid value within the specified bounds after max retries.",
                self.distribution,
                self.max_retries,
                self.lower_bound,
                self.upper_bound
            )
else:
    class Sampler(BaseSampler):
        """
        A class to sample from various statistical distributions using the standard random module.

        Note: The expovariate method has been adjusted to align with NumPy's exponential function,
        using the scale directly as the mean of the distribution.
        """

        def __init__(self, **kwargs):
            """
            Initialize the Sampler with standard random module-specific parameters.

            :param kwargs: Keyword arguments for initialization.
            """
            random_seed = kwargs.pop('random_seed', None)
            random_instance = kwargs.pop('random_instance', None)
            self.validate_random_parameters(random_seed, random_instance)

            super().__init__(**kwargs)

            if random_instance is not None:
                self.random_instance = random_instance
            else:
                self.random_instance = _random.Random(random_seed)

        def get_sample(self) -> Union[float, 'np.float32', 'np.float64']:
            """
            Get a sample from the specified distribution using the standard random module.

            :return: A sample from the specified distribution within the specified bounds.
            """
            logger.info("get_sample()")

            distribution_methods = {
                'lognormvariate': lambda: self.random_instance.lognormvariate(math.log(self.scale), self.shape),
                'normalvariate': lambda: self.random_instance.normalvariate(self.scale, self.shape),
                'expovariate': lambda: self.random_instance.expovariate(self.scale),
                'vonmisesvariate': lambda: self.random_instance.vonmisesvariate(self.scale, self.shape),
                'gammavariate': lambda: self.random_instance.gammavariate(self.shape, self.scale),
                'gauss': lambda: self.random_instance.gauss(self.scale, self.shape),
                'betavariate': lambda: self.random_instance.betavariate(self.shape, self.scale),
                'paretovariate': lambda: self.random_instance.paretovariate(self.shape),
                'weibullvariate': lambda: self.random_instance.weibullvariate(self.shape, self.scale)
            }

            for _ in range(self.max_retries):
                sampled_value = distribution_methods[self.distribution]()
                if self.lower_bound <= sampled_value <= self.upper_bound:
                    return sampled_value
            raise SamplingError(
                "Failed to sample a valid value within the specified bounds after max retries.",
                self.distribution,
                self.max_retries,
                self.lower_bound,
                self.upper_bound
            )


import logging
import pytest
import alexber.utils._ymlparsers_extra as ymlparsers_extra
from alexber.utils._ymlparsers_extra import _convert_template_to_string_format, \
    __convert_template_to_string_format as uuconvert_template_to_string_format, \
    format_template



try:
    from .ymlparsers_test import ymlparsersSetup, ymlparsersCleanup, ymlparsers, create_mock_lock
    from alexber.utils._ymlparsers_extra import HiYaPyCo

    from jinja2 import Environment as _Environment, \
                 DebugUndefined as _DebugUndefined
except ImportError:
    pass

logger = logging.getLogger(__name__)


@pytest.fixture
def ymlparsersExtraFixture(request, mocker, ymlparsersSetup, ymlparsersCleanup):
    p_isHiYaPyCoAvailable = ymlparsers_extra._isHiYaPyCoAvailable
    p_isJinja2DefaultAvailable = ymlparsers_extra._isJinja2DefaultAvailable

    if p_isHiYaPyCoAvailable and p_isJinja2DefaultAvailable:
        request_param = (p_isHiYaPyCoAvailable, p_isJinja2DefaultAvailable) if (not hasattr(request, 'param')) else request.param
    else:
        request_param = (False, False)
    isHiYaPyCoAvailable, isJinja2DefaultAvailable = request_param

    ymlparsers_extra._isHiYaPyCoAvailable = isHiYaPyCoAvailable
    ymlparsers_extra._isJinja2DefaultAvailable = isJinja2DefaultAvailable
    ymlparsers_extra._init_globals()

    yield request_param
    ymlparsers_extra._isHiYaPyCoAvailable = p_isHiYaPyCoAvailable
    ymlparsers_extra._isJinja2DefaultAvailable = p_isJinja2DefaultAvailable
    ymlparsers_extra._init_globals()



@pytest.mark.parametrize(
     'template, exp_value',

    [
        ('ping {{app.inner_host_name}}', 'ping {app_inner_host_name}'),
        ('Hi, {{user}}. My name is {{app.first.name}}. Good Luck!', 'Hi, {user}. My name is {app_first_name}. Good Luck!'),
        ('plain text', 'plain text'),
        ('First. Second.', 'First. Second.'),
        ('{{name}}', '{name}'),
        ('{{app,name}}', '{app,name}'),
        ('{{name}}.', '{name}.'),
        ('{{app,name}}.', '{app,name}.'),
        ('', ''),
         (None, None),


    ]
)

def test_uuconvert_template_to_string_format(request, mocker, template, exp_value):
    logger.info(f'{request._pyfuncitem.name}()')

    value = uuconvert_template_to_string_format(template,
                                                default_start='{{',
                                                default_end='}}'
                                                )
    pytest.assume(exp_value==value)



def test_uuconvert_template_to_string_format_undocumented1(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_value = 'ping {{app.inner_host_name}}'
    value = uuconvert_template_to_string_format(exp_value, default_start=None, default_end=None)
    pytest.assume(exp_value == value)

def test_uuconvert_template_to_string_format_undocumented2(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_value = 'ping {{app.inner_host_name}}'
    value = uuconvert_template_to_string_format(exp_value)
    pytest.assume(exp_value == value)


def test_uuconvert_is_used_in_uconvert_template_to_string_format(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    uconvert_mock=mocker.patch('alexber.utils._ymlparsers_extra.__convert_template_to_string_format',
                               side_effect=uuconvert_template_to_string_format, autospec=True, spec_set=True)

    exp_value = '{name}'
    template = '{{name}}'
    value = _convert_template_to_string_format(template)
    pytest.assume(exp_value == value)
    pytest.assume(uconvert_mock.call_count > 0)


#isHiYaPyCoAvailable,isJinja2DefaultAvailable
@pytest.mark.parametrize('ymlparsersExtraFixture', [(False, False), #minimal, no 3-rd party dependencies
                                                    (True, True),   #maximal 3-rd party dependencies available, \
                                                                    #but not customized
                                                    (False, True),  #only Jinja2 avaialble
                                                    (True, False),  #can't really happen
                                                    ], indirect=True)
def test_uconvert_template_to_string_format_minimal(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    exp_value = '{name}'
    template = '{{name}}'
    value = _convert_template_to_string_format(template)
    pytest.assume(exp_value == value)


#isHiYaPyCoAvailable,isJinja2DefaultAvailable
@pytest.mark.parametrize('ymlparsersExtraFixture', [(False, True)  #only Jinja2 avaialble \
                                                    ], indirect = True)
@pytest.mark.yml
def test_uconvert_template_to_string_jinja2DefaultChanged(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')


    mocker.patch('jinja2.defaults.VARIABLE_START_STRING', new='1_', spec_set=True)
    mocker.patch('jinja2.defaults.VARIABLE_END_STRING', new='_1', spec_set=True)

    ymlparsers_extra._init_globals()
    exp_value = '{name}'
    template = '1_name_1'
    value = _convert_template_to_string_format(template)
    pytest.assume(exp_value == value)

#isHiYaPyCoAvailable,isJinja2DefaultAvailable
@pytest.mark.parametrize('ymlparsersExtraFixture', [(True, False)
                                                    ], indirect = True)
@pytest.mark.yml
def test_uconvert_template_to_string_HiYaPyCoDefault(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    ymlparsers.initConfig()

    mock_lock = create_mock_lock(mocker)
    init_jinja2ctx = HiYaPyCo.jinja2ctx
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
    jinja2ctx_mock = mocker.patch.object(HiYaPyCo, 'jinja2ctx', spec_set=True)
    mock_variable_start_string = mocker.PropertyMock(return_value=init_jinja2ctx.variable_start_string)
    type(jinja2ctx_mock).variable_start_string = mock_variable_start_string
    mock_variable_end_string = mocker.PropertyMock(return_value=init_jinja2ctx.variable_end_string)
    type(jinja2ctx_mock).variable_end_string = mock_variable_end_string

    exp_value = '{name}'
    template = '{{name}}'
    value = _convert_template_to_string_format(template)
    pytest.assume(exp_value == value)

    pytest.assume(mock_lock.acquire.call_count > 0)
    pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

    pytest.assume(mock_variable_start_string.call_count > 0)
    pytest.assume(mock_variable_end_string.call_count > 0)


#isHiYaPyCoAvailable,isJinja2DefaultAvailable
@pytest.mark.parametrize('ymlparsersExtraFixture', [(True, False)
                                                    ], indirect = True)
@pytest.mark.yml
def test_uconvert_template_to_string_HiYaPyCoDefaultChanged(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')


    ymlparsers.initConfig(jinja2ctx={'variable_start_string':'2_', 'variable_end_string':'_2' })
    init_jinja2ctx = HiYaPyCo.jinja2ctx

    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
    jinja2ctx_mock = mocker.patch.object(HiYaPyCo, 'jinja2ctx', spec_set=True)
    mock_variable_start_string = mocker.PropertyMock(return_value=init_jinja2ctx.variable_start_string)
    type(jinja2ctx_mock).variable_start_string = mock_variable_start_string
    mock_variable_end_string = mocker.PropertyMock(return_value=init_jinja2ctx.variable_end_string)
    type(jinja2ctx_mock).variable_end_string = mock_variable_end_string

    exp_value = '{name}'
    template = '2_name_2'
    value = _convert_template_to_string_format(template)
    pytest.assume(exp_value == value)

    pytest.assume(mock_lock.acquire.call_count > 0)
    pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

    pytest.assume(mock_variable_start_string.call_count > 0)
    pytest.assume(mock_variable_end_string.call_count > 0)

@pytest.mark.yml
def test_uconvert_template_to_string_explicit_param1_jinja2ctx(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    mock_lock = create_mock_lock(mocker)

    exp_value = '{name}'
    template = '3_name_3'

    jinja2ctx = _Environment(undefined=_DebugUndefined, variable_start_string='3_', variable_end_string='_3')
    value = _convert_template_to_string_format(template, jinja2ctx=jinja2ctx)
    pytest.assume(exp_value == value)

@pytest.mark.yml
def test_uconvert_template_to_string_explicit_param1_jinja2Lock(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)

    #exp_value = '{name}'
    template = '{{name}}'   #HiYaPyCo.jinja2ctx is default one

    jinja2Lock_param_lock = create_mock_lock(mocker)
    with pytest.raises(ValueError):
        _convert_template_to_string_format(template, jinja2Lock=jinja2Lock_param_lock)
    #pytest.assume(exp_value == value)
    pytest.assume(mock_lock.acquire.call_count == 0)
    #pytest.assume(jinja2Lock_param_lock.acquire.call_count ==0)



#isHiYaPyCoAvailable,isJinja2DefaultAvailable
@pytest.mark.parametrize('ymlparsersExtraFixture', [(False, True)   #only Jinja2 avaialble \
                                                    ], indirect = True)
@pytest.mark.yml
def test_uconvert_template_to_string_explicit_param1a_jinja2Lock(request, mocker, ymlparsersExtraFixture,
                                                                ):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
    jinja2ctx_mock = mocker.patch.object(HiYaPyCo, 'jinja2ctx', spec_set=True)

    exp_value = '{name}'
    template = '{{name}}'   #isHiYaPyCoAvailable is not available, global default are used and external lock
                            # (will be ignored)

    jinja2Lock_param_lock = create_mock_lock(mocker)
    value = _convert_template_to_string_format(template, jinja2Lock=jinja2Lock_param_lock)
    pytest.assume(exp_value == value)
    pytest.assume(mock_lock.acquire.call_count == 0)
    pytest.assume(jinja2ctx_mock.call_count == 0)

@pytest.mark.yml
def test_uconvert_template_to_string_explicit_param1b_jinja2Lock(request, mocker, ymlparsersExtraFixture,
                                                               ):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    ymlparsers.initConfig(jinja2ctx={'variable_start_string': '22_', 'variable_end_string': '_22'})

    init_jinja2ctx = HiYaPyCo.jinja2ctx
    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
    jinja2ctx_mock = mocker.patch.object(HiYaPyCo, 'jinja2ctx', spec_set=True)
    mock_variable_start_string = mocker.PropertyMock(return_value=init_jinja2ctx.variable_start_string)
    type(jinja2ctx_mock).variable_start_string = mock_variable_start_string
    mock_variable_end_string = mocker.PropertyMock(return_value=init_jinja2ctx.variable_end_string)
    type(jinja2ctx_mock).variable_end_string = mock_variable_end_string

    #exp_value = '{name}'
    template = '22_name_22'   #HiYaPyCo.jinja2ctx is non default one and external lock

    jinja2Lock_param_lock = create_mock_lock(mocker)
    with pytest.raises(ValueError):
        _convert_template_to_string_format(template, jinja2Lock=jinja2Lock_param_lock)
    #pytest.assume(exp_value == value)
    pytest.assume(mock_lock.acquire.call_count == 0)
    #pytest.assume(jinja2Lock_param_lock.acquire.call_count ==0)


#isHiYaPyCoAvailable,isJinja2DefaultAvailable
@pytest.mark.parametrize('ymlparsersExtraFixture', [(False, True)
                                                    ], indirect = True)
@pytest.mark.yml
def test_uconvert_template_to_string_explicit_param1a_jinja2ctx(request, mocker, ymlparsersExtraFixture,
                                                               ):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)

    exp_value = '{name}'
    template = '33_name_33'

    jinja2ctx = _Environment(undefined=_DebugUndefined, variable_start_string='33_', variable_end_string='_33')
    value = _convert_template_to_string_format(template, jinja2ctx=jinja2ctx)

    pytest.assume(exp_value == value)
    pytest.assume(mock_lock.acquire.call_count == 0)


@pytest.mark.yml
def test_uconvert_template_to_string_explicit_param2(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}{ymlparsersExtraFixture}')

    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
    mock_lock_inuse = create_mock_lock(mocker)

    exp_value = '{name}'
    template = '4_name_4'

    jinja2ctx = _Environment(undefined=_DebugUndefined, variable_start_string='4_', variable_end_string='_4')
    value = _convert_template_to_string_format(template, jinja2ctx=jinja2ctx, jinja2Lock=mock_lock_inuse)
    pytest.assume(exp_value == value)

    pytest.assume(mock_lock.acquire.call_count == 0)
    pytest.assume(mock_lock_inuse.acquire.call_count >0)
    pytest.assume(mock_lock_inuse.release.call_count == mock_lock_inuse.acquire.call_count)


def test_format_template(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    uconvert_mock=mocker.patch('alexber.utils._ymlparsers_extra._convert_template_to_string_format',
                               side_effect=_convert_template_to_string_format, autospec=True, spec_set=True)

    exp_value = 'Hello, John!'
    template = 'Hello, {{name}}!'
    value = format_template(template, name='John')
    pytest.assume(exp_value == value)
    pytest.assume(uconvert_mock.call_count > 0)

def test_format_template_without_variables(request, mocker, ymlparsersExtraFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    uconvert_mock=mocker.patch('alexber.utils._ymlparsers_extra._convert_template_to_string_format',
                               side_effect=_convert_template_to_string_format, autospec=True, spec_set=True)

    exp_value = 'Hello, World!'
    template = 'Hello, World!'
    value = format_template(template)
    pytest.assume(exp_value == value)
    pytest.assume(uconvert_mock.call_count > 0)


if __name__ == "__main__":
    pytest.main([__file__])



import logging

import shutil
from pathlib import Path

logger = logging.getLogger(__name__)


def join_files(f):
    """
    We will read all files that start from the same file name,
    and will write it's content to the specified file.

    :param f: output file, also pattern for other files to look on.
    :return:
    """
    output_file = Path(str(f))
    simple_filename = output_file.stem
    suffix = output_file.suffix
    p = output_file.parent

    gen = p.glob(f"{simple_filename}_*{suffix}")
    # see https://stackoverflow.com/a/27077437/1137529
    with open(str(output_file), 'w') as wfd:
        while True:
            try:
                partial_output = next(gen)
                with open(str(partial_output), 'r') as fd:
                    shutil.copyfileobj(fd, wfd)
            except StopIteration as e:
                break



import pstats
import sys

#See https://alex-ber.medium.com/converting-prof-files-to-callgrind-f6ae56fae0d3 for more details.
def convert_prof_to_callgrind(input_file, output_file):
    # Load the profile statistics
    stats = pstats.Stats(input_file)

    with open(output_file, 'w') as f:
        f.write("events: ns\n")  # Use nanoseconds as the unit

        # Write function and caller statistics to the callgrind file
        for func, (cc, nc, tt, ct, callers) in stats.stats.items():
            filename, line, name = func
            f.write(f"fl={filename}\n")
            f.write(f"fn={name}\n")
            f.write(f"{line} {ct}\n")  # Time is in the original units

            for caller, (cc, nc, tt, ct) in callers.items():
                c_filename, c_line, c_name = caller
                f.write(f"cfl={c_filename}\n")
                f.write(f"cfn={c_name}\n")
                f.write(f"calls={nc} {c_line}\n")
                f.write(f"{line} {ct}\n")  # Time is in the original units

            f.write("\n")


def main():
    if len(sys.argv) != 3:
        print("Usage: python prof_to_callgrind.py input.prof output.callgrind")
        sys.exit(1)

    input_file = sys.argv[1]
    output_file = sys.argv[2]

    convert_prof_to_callgrind(input_file, output_file)


if __name__ == "__main__":
    main()

# Usage note:
# The convert_prof_to_callgrind function converts .prof files to .callgrind format.
# To use the script, run: python prof_to_callgrind.py input.prof output.callgrind


"""
See https://medium.com/analytics-vidhya/my-parser-module-429ed1457718 for documentation.
"""
import logging
logger = logging.getLogger(__name__)
from collections import OrderedDict

from configparser import ConfigParser
from argparse import ArgumentParser
import sys
import ast


def _as_dict(parser):
    """
    Go over all sections in the parser,
    convert's there's key/value as dictionary that
    for every section in the parser has dictionary
    with key/value pairs (both str).

    :param parser: (self)
    :return: dict with key/value as str
    """
    d = OrderedDict()
    for section in parser.sections():
        d[section] = OrderedDict()
        for key in parser.options(section):
            d[section][key] = parser.get(section, key)
    return d

ConfigParser.as_dict = _as_dict

#inspired by
#https://stackoverflow.com/questions/21920989/parse-non-pre-defined-argument
#https://stackoverflow.com/questions/51267814/argparse-for-unknown-number-of-arguments-and-unknown-names
def _args_as_dict(parser, args=None):
    """
    Go over source for arguments, takes argument of the form --key=value.
    Create dictionary.
    Strip out '--' prefix from the key and put key/value (as str) to dict.

    If args is None, sys.argv[1:] will be used as source for arguments.
    Note: sys.argv[0] is ignored as it contain the name of main .py file to run.

    :param parser: ArgumentParser (self)
    :param args: if is not None, will be used as source for arguments.
    :return:
    """

    #see #argumentParser.parse_known_args()
    if args is None:
        # args default to the system args
        args = sys.argv[1:]
    else:
        # make sure that args are mutable
        args = list(args)

    d = OrderedDict()

    key = None
    value = None
    for arg in args:
        if arg.startswith('--') and '=' in arg:
            key, value = arg.rsplit("=", 1)
        else:
            key = arg
            value = None
        if key.startswith('--'):
            key = key[2:]

        d[key] = value

    return d


ArgumentParser.as_dict = _args_as_dict


#insipred by https://stackoverflow.com/a/14258151/1137529
#
def safe_eval(value):
    '''
    The purpose of this function is convert numbers from str to correct type.
    This function support convertion of built-in Python number to correct type (int, float)
    This function doesn't support decimal.Decimal or datetime.datetime or numpy types.
    '''
    try:
        ret = ast.literal_eval(value)
    except (SyntaxError, ValueError):
        ret = value
    return ret

def is_empty(value):
    '''
    if value is None returns True.

    if value is empty iterable (for example, empty str or emptry list),
    returns true
    otherwise false

    Note: For not iterable values, behaivour is undefined.

    :param value:
    :return:
    '''
    if value is None:
        return True
    if value:
        ret = False
    else:
        ret = True
    return ret


def parse_boolean(value):
    '''
    if value is None returns None.

    if value is boolean, it is returned as it is.
    if value is str and value is equals ignoring case to "True", True is returned.
    if value is str and value is equals ignoring case to "False", False is returned.

    For every other value, the answer is undefined.

    :param value:
    :return:
    '''
    if value is None:
        return None

    if value in (True, False):
        return value
    try:
        return {"true": True, "false": False}[value.casefold()]
    except (AttributeError, KeyError):
        raise ValueError(f"unknown string for bool: {value!r}")

def parse_sys_args(argumentParser=None, args=None):
    """
    This function parses command line arguments.

    :param argumentParser:
    :param args: if not None, suppresses sys.args
    :return:
    """
    if argumentParser is None:
        argumentParser = ArgumentParser()
    argumentParser.add_argument("--general.config.file", nargs='?', dest='config_file', default='config.yml',
                                const='config.yml')
    params, unknown_arg = argumentParser.parse_known_args(args=args)

    sys_d = argumentParser.as_dict(args=unknown_arg)
    return params, sys_d
